import json
from argparse import Namespace
from functools import partial
from pathlib import Path
from typing import NotRequired, Optional, TypedDict, Union

import pandas as pd
import surface_analyses.commandline_electrostatic
import surface_analyses.commandline_hydrophobic
from Bio.PDB import PDBParser, Structure
from feature_generation import (
    compute_surface_integrals,
    get_feat_name,
    get_global_electrostatic_descriptors_for_patches,
    get_global_hydrophobic_descriptors,
    get_global_hydrophobic_descriptors_for_patches,
    get_vertices_sub_dfs,
)
from get_atom_buriedness import get_atom_buriedness
from peppatch_postprocess import assign_patches_electrostatic, get_vertex_df
from surface_analyses.patches import assign_patches
from surface_analyses.structure import load_trajectory_using_commandline_args
from surface_analyses.surface import Surface


class ElectrostaticPatchConfig(TypedDict):
    patch_cutoff: tuple[float]  # (2., -2.)  default setting in pep patch
    name: str  # identifier/description
    aggregate_per_residue: NotRequired[bool]


class HydrophobicPatchConfig(TypedDict):
    patch_min: float  # threshold value (potential) for inclusion of a vertex in a patch, default of pep patch: 0.12 (works well with WiWh)    name: str
    name: str
    aggregate_per_residue: NotRequired[bool]


class ElectrostaticConfig(TypedDict):
    pH: float  # e.g. 7.4 of blood
    volume_integral_cutoff: NotRequired[tuple[float]]  # (0.3, -0.3)  default setting in pep patch
    patch_configs: list[ElectrostaticPatchConfig]

    # just for computing ML features:
    surface_integral_high_cutoffs: list[float]
    surface_integral_low_cutoffs: list[float]


class HydrophobicConfig(TypedDict):
    hydrophobicity_scale: str  # path to scale, ex: data/hydrophobicity_scales/scale_WiWh.csv
    patch_configs: list[HydrophobicPatchConfig]

    # just for computing ML features:
    surface_integral_cutoffs: list[float]


class PepPatchMethod(TypedDict):
    name: str  # identifier/description e.g. electrostatic_ph7.4
    method: str  # electrostatic/hydrophobic
    config: Union[ElectrostaticConfig, HydrophobicConfig]  # pH, hydrophobicity scale


class PepPatchSet(TypedDict):
    buriedness_radii: list[float]
    methods: list[PepPatchMethod]


class EventType(TypedDict):
    sample_id: str
    structure_prediction_method: str
    structure_prediction_scheme: str

    set_name: str
    set: NotRequired[
        PepPatchSet
    ]  # if not specified, hardcoded settings from codebase will be selected based on 'set_name'


def get_atom_to_residue_mapping(structure: Structure, atom_buriedness: pd.DataFrame = None):
    """Return DataFrame with atom index (0-based contiguous) -> chain_id, residue_id, residue_serial, atom_number (in pdb records)

    residue_id is '80' or '80A', residue_serial is zero-based and starts with H and then L chain residues (if not VHH).

    Residue serial is helpful when computing the CDRS using extended pep-patch code.
    """
    if atom_buriedness is None:
        atom_buriedness = get_atom_buriedness(structure, [0.0])  # 0.0 just as a placeholder

    return atom_buriedness[["residue_id", "chain_id", "residue_serial", "atom_number"]]


def patch_membership_agg(sub_df, patch_col):
    """Get the area of each patch on a residue (or a groupby). Performs an aggregation within a residue group (meant
      to by `apply`ed to a DataFrameGroupBy) to sum vertex 'area' for each value of patch_id (patch_col).
    Returns a dictionary with patch_id: respective area.  ex: {0: 0.7975, 3: 0.7426}
    """
    return sub_df.groupby(patch_col)["area"].sum().to_dict()


def agg_patches_per_residue(patches_vertices):
    """Convert patch vertices into residue-level patches.

    Returns a DataFrame, ex:
                              area      value                                       patch_labels  patch_label
    chain_id residue_id
    H        1           23.310820   2.124569                            {1: 23.310819893187727}            1
             13           9.196304   2.235697                             {19: 9.19630379374818}           19

    Returns a DataFrame with residue SASA, mean vertex value, and patches per residue and the patch with max area on
    the residue.
    """

    patches_gb_residue = patches_vertices.groupby(["chain_id", "residue_id"])

    residue_patches = patches_gb_residue.agg(
        {
            "area": "sum",
            "value": "mean",
            # todo atom buriedness?? of the patch? for position-ml feats
        }
    )

    # all patches the residue belongs in + respective area on the residue (dict path_id: area)
    residue_patches["patch_labels"] = patches_gb_residue.apply(
        partial(patch_membership_agg, patch_col="patch_id"), include_groups=False
    )
    # patch_id of the largest patch
    residue_patches["patch_label"] = residue_patches["patch_labels"].apply(lambda x: max(x, key=x.get))
    return residue_patches


def run_electrostatic(pdb_file: str, config: ElectrostaticConfig, atom_to_residue: pd.DataFrame, sequence: str):
    """Run PEP-Patch electrostatic and return surface vertices annotated with patch labels for each patch cutoff.

    Returns PEP-Patch result an an iterator of tuples patch_config, and DataFrame patch_vertices.

              area  atom   cdr     value  patch_id  residue_id    chain_id
    0     0.144793   410  True -0.226922         0          29           H
    1     0.084391   410  True -0.224813         0          29           H
    2     0.100023   410  True -0.211886         0          29           H

    PEP Patch electrostatic runtime on mac: 10.5s, then 0.15s/(one patch cutoff) takes the agg in patches and per-residue

    For patch_vertices see ` surf_to_patch_vertices_electrostatic`. `patch_id` is 0 if a patch is not assigned,
        negative integer for negative patch and positive int for positive patch
    """
    # pH (not patch_cutoff) determines the output surface potential of PEP-Patch
    mock_trajectory_args = Namespace(parm=pdb_file, trajs=[pdb_file], stride=1, ref=None, protein_ref=None)
    traj = load_trajectory_using_commandline_args(mock_trajectory_args)

    pep_result = surface_analyses.commandline_electrostatic.run_electrostatics(
        traj,
        apbs_dir="apbs",
        pH=config["pH"],
        out="/dev/null",  # we are using the return value in `result`
    )

    # annotated vertices (cdrs, potential value, closest atom number, ...)
    vertex_data = get_vertex_df(pep_result["surface"], atom_to_residue, sequence, annotate_cdrs=False)
    # add residue id to each vertex (based on atom number)
    vertex_data = vertex_data.merge(atom_to_residue, left_on="atom", right_index=True)
    faces = pep_result["surface"].faces
    # todo could save (vertex_data, faces); just from this, patches and features are computed, maybe also `sequence` (or seqH + seqL)

    # return patches for every patch_cutoff
    def patch_vertices_iterator():
        for patch_config in config["patch_configs"]:
            patch_vertices = vertex_data.copy()
            patch_vertices["patch_id"] = assign_patches_electrostatic(
                vertex_data["value"], faces, patch_config["patch_cutoff"]
            )
            yield patch_config, patch_vertices

    return pep_result, patch_vertices_iterator()


def run_hydrophobic(pdb_file: str, config: HydrophobicConfig, atom_to_residue: pd.DataFrame, sequence: str):
    """Run PEP-Patch hydrophobic and return surface vertices annotated with patch labels for each patch cutoff.

    Returns an iterator of tuples patch_cutoff, and DataFrame patch_vertices.

              area  atom   cdr     value  patch_id  residue_id    chain_id
    0     0.144793   410  True -0.226922         0          29           H
    1     0.084391   410  True -0.224813         0          29           H
    2     0.100023   410  True -0.211886         0          29           H

    PEP hydrophobic runtime on mac: 3s, then 0.25s/(one patch cutoff) takes the agg in patches and per-residue

    Residues with hydrophobic patches have positive patch label (1..inf)
    Residues not belonging to a patch have a label 0 (if the residue vertices are mostly unassigned to a patch).
    """
    # hydrophobicity_scale determines the output surface potential of PEP-Patch (not patch_min's)
    mock_trajectory_args = Namespace(parm=pdb_file, trajs=[pdb_file], stride=1, ref=None, protein_ref=None)
    traj = load_trajectory_using_commandline_args(mock_trajectory_args)
    pep_result = surface_analyses.commandline_hydrophobic.run_hydrophobic(
        pdb_file,
        traj,
        scale=str(config["hydrophobicity_scale"]),
        potential=True,
        sap=True,
        surfscore=True,
        sh=True,  # additional per-atom features computed
    )

    surf = Surface.from_dict(pep_result, "hydrophobic_potential:0")
    # we could save surf

    # annotated vertices (cdrs, potential value, closest atom number, ...)
    vertex_data = get_vertex_df(surf, atom_to_residue, sequence, annotate_cdrs=False)
    vertex_data["area"] *= (
        100  # for some reason, pep_patch_hydrophobic computes with nanometers, but pep_patch_electrostatic with angstroms, convert nanometers^2 to angstroms^2
    )
    # add residue id to each vertex (based on atom number)
    vertex_data = vertex_data.merge(atom_to_residue, left_on="atom", right_index=True)
    faces = surf.faces
    # todo could save (vertex_data, faces); just from this, patches and features are computed, maybe also `sequence` (or seqH + seqL)

    # return patches for every patch_min
    def patch_vertices_iterator():
        for patch_config in config["patch_configs"]:
            patch_vertices = vertex_data.copy()
            patch_vertices["patch_id"] = assign_patches(faces, vertex_data["value"] > patch_config["patch_min"])
            patch_vertices["patch_id"] += (
                1  # 0 is not assigned, 1..inf are patches (for consistency with electrostatic)
            )
            yield patch_config, patch_vertices

    return pep_result, patch_vertices_iterator()


def process_pdb(input_pdb: str, peppatch_set: PepPatchSet):
    pdb_structure = PDBParser().get_structure("structure", input_pdb)

    sequence: str = ""  # Not used (logic for antibody CDR annotation)
    atom_buriedness = (
        get_atom_buriedness(pdb_structure, peppatch_set["buriedness_radii"])
        if peppatch_set["buriedness_radii"]
        else None
    )

    # get mapping atom_index -> chain_id, residue_id, so we can map atom numbers
    #   returned from PEP-Patch to residues on a chain
    atom_to_residue = get_atom_to_residue_mapping(pdb_structure, atom_buriedness)

    mab_type = ""  # Not used (logic for antibody CDR annotation)

    # function to aggregate patches per residue for visualization
    per_residue_patches = []

    def maybe_agg_per_residue(method, patch_config, patch_vertices):
        if "aggregate_per_residue" not in patch_config or not patch_config["aggregate_per_residue"]:
            return
        per_residue_patches.append(
            dict(
                **method,  # to identify the parameters the result below was computed with
                patch_config=patch_config,
                per_residue_patches=agg_patches_per_residue(patch_vertices).reset_index().to_dict(orient="records"),
            )
        )

    global_descriptors = {}
    for method in peppatch_set["methods"]:
        method_name = method["name"]

        # assert there are non-empty patch configs (patch_min)
        assert method["config"]["patch_configs"]

        def add_descriptors(
            descriptors: dict[str, Union[float, int]],
            patch_cutoff: Optional[Union[float, tuple[float]]] = None,
            sub_df_prefix: str = "",
        ):
            global_descriptors.update(
                {
                    get_feat_name(method_name, k, patch_cutoff, sub_df_prefix, mab_type): v
                    for k, v in descriptors.items()
                }
            )

        if method["method"] == "electrostatic":
            pep_result, patch_vertices_it = run_electrostatic(input_pdb, method["config"], atom_to_residue, sequence)

            # volume potential integrals computed by pep_patch_electrostatic
            add_descriptors(pep_result["integrals"])

            for patch_config, patch_vertices in patch_vertices_it:
                # optional per-residue aggregation for visualization in biophi
                maybe_agg_per_residue(method, patch_config, patch_vertices)
                # patch-based global descriptors
                add_descriptors(
                    get_global_electrostatic_descriptors_for_patches(patch_vertices, atom_buriedness),
                    patch_config["patch_cutoff"],
                )

            # surface integrals
            # (use last patch_vertices, as only patch definitions differ in the loop above, but a bit hacky)
            for sub_df_name, vertex_df in get_vertices_sub_dfs(patch_vertices):
                # this descriptor is in fact total patch area by the region (H/L cdr1 etc..), weighted by the intensity
                add_descriptors(
                    compute_surface_integrals(
                        vertex_df,
                        high_cutoffs=method["config"]["surface_integral_high_cutoffs"],
                        low_cutoffs=method["config"]["surface_integral_low_cutoffs"],
                    ),
                    sub_df_prefix=sub_df_name,
                )
        elif method["method"] == "hydrophobic":
            # elif method['method'] == 'hydrophobic':
            pep_result, patch_vertices_it = run_hydrophobic(input_pdb, method["config"], atom_to_residue, sequence)

            # additional descriptors computed by pep_patch_hydrophobic
            # e.g. surfscore is defined as saa * propensity of each atom, summed for the structure
            add_descriptors(get_global_hydrophobic_descriptors(pep_result))

            for patch_config, patch_vertices in patch_vertices_it:
                # optional per-residue aggregation for visualization in biophi
                maybe_agg_per_residue(method, patch_config, patch_vertices)

                # patch-based global descriptors
                add_descriptors(
                    get_global_hydrophobic_descriptors_for_patches(patch_vertices, atom_buriedness),
                    patch_config["patch_min"],
                )

            # surface integrals
            # (use last patch_vertices, as only patch definitions differ in the loop above, but a bit hacky)
            for sub_df_name, vertex_df in get_vertices_sub_dfs(patch_vertices):
                add_descriptors(
                    compute_surface_integrals(vertex_df, high_cutoffs=method["config"]["surface_integral_cutoffs"]),
                    sub_df_prefix=sub_df_name,
                )
        else:
            raise ValueError(f"Method {method['method']} not supported")

    if peppatch_set["methods"]:
        # reuse last variable - patch_vertices - for global definitions of cdr123 and chain total surface
        #   areas for the structure (only depends on the predicted structure and no other params)
        for sub_df_name, vertex_df in get_vertices_sub_dfs(patch_vertices):
            global_descriptors[f"{sub_df_name}area"] = vertex_df.area.sum()

    return per_residue_patches, global_descriptors


def get_peppatch_set_definitions(sets_path: str):
    """Our PEP-Patch settings for computing features/residue annotations.

    Can be stored here in this repo for versioning/consistency.
    """
    with Path(sets_path).open() as f:
        return json.load(f)


# CONFIG_STRUCTURE_BUCKET = os.getenv('CONFIG_STRUCTURE_BUCKET')
# CONFIG_DESCRIPTOR_BUCKET = os.getenv('CONFIG_DESCRIPTOR_BUCKET')


# def lambda_handler(event: EventType, context):
#     print('Received event:', event)
#     sample_id = event['sample_id']
#     set_name = event['set_name']
#     input_pdb_path_s3 = f'{sample_id}/{event["structure_prediction_method"]}_{event["structure_prediction_scheme"]}.pdb'
#     output_path_s3_base = f'{sample_id}/{event["structure_prediction_method"].lower()}_{set_name}'

#     try:
#         # get the `methods` from set definition, if there is one for the set_name, otherwise use event['methods']
#         set_definitions = get_peppatch_set_definitions()
#         if set_name in set_definitions:
#             assert 'methods' not in event, f'Set {set_name} methods are already defined in `data/feature_sets.json` but received another `methods` definitions in the Event.'
#             peppatch_set = set_definitions[set_name]
#         else:
#             peppatch_set = event['set']


#         # Use temp dir as working directory,
#         with tempfile.TemporaryDirectory() as tmpdirname, contextlib.chdir(tmpdirname):
#             input_pdb_filename = 'input.pdb'
#             s3_client.download_file(CONFIG_STRUCTURE_BUCKET, input_pdb_path_s3, input_pdb_filename)

#             per_residue_patches, global_descriptors = process_pdb(input_pdb_filename, peppatch_set)

#         s3_client.put_object(
#             Bucket=CONFIG_DESCRIPTOR_BUCKET,
#             Key=f'{output_path_s3_base}.csv',
#             Body=pd.DataFrame([global_descriptors]).to_csv(index=False).encode('utf-8'),
#         )

#         s3_client.put_object(
#             Bucket=CONFIG_DESCRIPTOR_BUCKET,
#             Key=f'{output_path_s3_base}_per_residue_patches.json',  # todo user cannot change numbering scheme after upload, no? If yes, include it here in the name
#             Body=json.dumps(per_residue_patches).encode('utf-8'),
#         )

#     except Exception as e:
#         traceback.print_exc()

#         s3_client.put_object(
#             Bucket=CONFIG_DESCRIPTOR_BUCKET,
#             Key=f'{output_path_s3_base}.err.json',
#             Body=json.dumps({'message': str(e), 'traceback': ''.join(traceback.format_exception(e))}).encode('utf-8'),
#         )



import enum
import json
from dataclasses import is_dataclass, asdict
from pathlib import Path

from sqlalchemy import TypeDecorator, String


class DataclassEncoder(json.JSONEncoder):
    """Simple JSON encoder that encodes dataclasses and enums."""

    def default(self, o):
        if hasattr(o, "__do_not_serialize__") and o.__do_not_serialize__:
            raise ValueError(f"Objects of type {type(o).__name__} cannot be updated in the database.")
        if is_dataclass(o):
            return asdict(o)
        if isinstance(o, enum.Enum):
            return o.value
        if isinstance(o, Path):
            return str(o.resolve())
        return super().default(o)


class DataclassType(TypeDecorator):
    """SQLA Type decorator to serialize dataclasses"""

    impl = String
    cache_ok = True

    def __init__(self, base_cls):
        super().__init__()
        self.base_cls = base_cls

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return json.dumps(value, cls=DataclassEncoder)

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        result = json.loads(value)
        if hasattr(self.base_cls, "from_dict"):
            return self.base_cls.from_dict(result)
        return self.base_cls(**result)



import shutil
import zipfile
import io
import pickle
from typing import List

from urllib.parse import urlparse
from botocore.exceptions import ClientError
from concurrent.futures import ThreadPoolExecutor

import os
import tempfile
import weakref
from collections import OrderedDict
from ovo.core.aws import AWSSessionManager
from ovo.core.utils.formatting import get_hashed_path_for_bytes


class Storage:
    """Class for managing file storage in local filesystem or S3 bucket"""

    def __init__(
        self,
        storage_root: str,
        aws: AWSSessionManager | None,
        verbose: bool = False,
        memory_cache_limit_bytes=50 * 1024 * 1024,
        disk_cache_limit_bytes=200 * 1024 * 1024,
        memory_cache_limit_per_file_bytes=5 * 1024 * 1024,
    ):
        self.storage_root: str = storage_root
        self.aws: AWSSessionManager | None = aws
        self.verbose = verbose
        # caching
        self.memory_cache_limit = memory_cache_limit_bytes
        self.disk_cache_limit = disk_cache_limit_bytes
        self.memory_cache_limit_per_file_bytes = memory_cache_limit_per_file_bytes
        self._cache_memory = OrderedDict()  # LRU cache for memory
        self._cache_disk = OrderedDict()  # LRU cache for disk
        self.memory_cache_size = 0
        self.disk_cache_size = 0
        self._temp_dir = tempfile.TemporaryDirectory()
        weakref.finalize(self, self._clear_temp_dir)

    def _clear_temp_dir(self):
        self._temp_dir.cleanup()

    def _cache_read(self, file_path):
        if file_path in self._cache_memory:
            self._cache_memory.move_to_end(file_path)  # Mark as recently used
            return self._cache_memory[file_path]
        if file_path in self._cache_disk:
            with open(self._cache_disk[file_path], "rb") as f:
                content = f.read()
            self._cache_memory[file_path] = content
            self.memory_cache_size += len(content)
            while self.memory_cache_size > self.memory_cache_limit and self._cache_memory:
                old_file, old_content = self._cache_memory.popitem(last=False)
                self.memory_cache_size -= len(old_content)
            return content
        return None

    def _cache_store(self, file_path, content: str | bytes):
        if len(content) <= self.memory_cache_limit_per_file_bytes:
            # Cache recent files in memory (unless they are too big)
            self._cache_memory[file_path] = content
            self.memory_cache_size += len(content)
            while self.memory_cache_size > self.memory_cache_limit and self._cache_memory:
                old_file, old_content = self._cache_memory.popitem(last=False)
                self.memory_cache_size -= len(old_content)
        scheme, netloc, relative_path = self.parse_path(file_path)
        if not scheme:
            # Do not use disk cache for files that are already local
            return
        assert relative_path in os.path.abspath(relative_path), (
            f"Invalid path {relative_path}"
        )  # make sure path doesn't contain '../'
        disk_file_path = os.path.join(self._temp_dir.name, relative_path)
        os.makedirs(os.path.dirname(disk_file_path), exist_ok=True)
        with open(disk_file_path, "wb") as f:
            f.write(content)
        self._cache_disk[file_path] = disk_file_path
        self.disk_cache_size += os.path.getsize(disk_file_path)
        while self.disk_cache_size > self.disk_cache_limit and self._cache_disk:
            old_file, old_disk_path = self._cache_disk.popitem(last=False)
            self.disk_cache_size -= os.path.getsize(old_disk_path)
            os.remove(old_disk_path)

    @staticmethod
    def parse_path(path: str) -> tuple[str, str, str]:
        parsed = urlparse(str(path), allow_fragments=False)
        return parsed.scheme, parsed.netloc, parsed.path.lstrip("/")

    def list_dir(self, abs_path: str, only_dir=False, recursive=False) -> list[str]:
        """List all files in the directory, return list of paths relative to provided path
        :param abs_path: path to the source directory
        :param only_dir: if True, only list directories, if False, list all files
        :param recursive: if True, list all files in the directory recursively (relative to provided path)
        :return: list of files in the directory
        """
        abs_path = self.resolve_path(abs_path)

        scheme, bucket, prefix = self.parse_path(abs_path)
        if scheme == "s3":
            prefix = prefix.rstrip("/") + "/"
            objects = []

            if recursive:
                paginator = self.aws.s3.get_paginator("list_objects_v2")
                for page in paginator.paginate(Bucket=bucket, Prefix=prefix):
                    if "CommonPrefixes" in page and only_dir:
                        objects += [cp["Prefix"][len(prefix) : -1] for cp in page["CommonPrefixes"]]
                    if "Contents" in page and not only_dir:
                        objects += [obj["Key"][len(prefix) :] for obj in page["Contents"] if obj["Key"] != prefix]
            else:
                try:
                    response = self.aws.s3.list_objects_v2(
                        Bucket=bucket, Prefix=prefix.rstrip("/") + "/", Delimiter="/"
                    )
                except ClientError as e:
                    print(e)
                    raise

                if "CommonPrefixes" in response:
                    objects += [
                        os.path.basename(content["Prefix"].rstrip("/"))
                        for content in response["CommonPrefixes"]
                        if content["Prefix"] != prefix
                    ]

                if "Contents" in response and not only_dir:
                    objects += [
                        os.path.basename(content["Key"].rstrip("/"))
                        for content in response["Contents"]
                        if content["Key"] != prefix
                    ]

            return objects

        if not os.path.exists(abs_path):
            return []

        if recursive:
            if only_dir:
                return sorted(
                    os.path.join(root, d).removeprefix(abs_path.rstrip("/") + "/")
                    for root, dirs, files in os.walk(abs_path, followlinks=True)
                    for d in dirs
                )
            else:
                return sorted(
                    os.path.join(root, filename).removeprefix(abs_path.rstrip("/") + "/")
                    for root, dirs, files in os.walk(abs_path, followlinks=True)
                    for filename in files
                )
        else:
            if only_dir:
                return sorted([d for d in os.listdir(abs_path) if os.path.isdir(os.path.join(abs_path, d))])
            else:
                return sorted(os.listdir(abs_path))

    def file_exists(self, storage_path) -> bool:
        """Search for the file in the local filesystem or in the S3 bucket
        :param storage_path: relative (within storage root) or absolute path to the source file
        :return: True if the file exists, False otherwise
        """
        abs_path = self.resolve_path(storage_path)

        scheme, bucket, key = self.parse_path(abs_path)
        if scheme == "s3":
            try:
                self.aws.s3.head_object(Bucket=bucket, Key=key)
                return True
            except ClientError as e:
                # Check if the error is "Not Found"
                if e.response["Error"]["Code"] in ("NoSuchKey", "404", 404):
                    return False
                else:
                    print("S3 ERROR", e, e.response)
                    raise e
        return os.path.exists(abs_path)

    def read_file_bytes(self, storage_path, cache_store: bool = True) -> bytes:
        abs_path = self.resolve_path(storage_path)

        # Check if the file is in memory cache
        content = self._cache_read(abs_path)
        if content is not None:
            if self.verbose:
                print(f"Reading {storage_path} from cache")
            return content

        # Read file contents from s3 or disk
        scheme, bucket, key = self.parse_path(abs_path)
        if scheme == "s3":
            try:
                if self.verbose:
                    print(f"Reading {storage_path} from s3")
                response = self.aws.s3.get_object(Bucket=bucket, Key=key)
                content = response["Body"].read()
                if cache_store:
                    self._cache_store(abs_path, content)
            except ClientError as e:
                print("S3 ERROR", e, e.response)
                if e.response["Error"]["Code"] in ("NoSuchKey", "404", 404):
                    raise FileNotFoundError(f"File not found: {abs_path}") from e
                raise
            return content

        if self.verbose:
            print(f"Reading {storage_path} from FS")
        with open(abs_path, "rb") as f:
            content = f.read()
        if cache_store:
            self._cache_store(abs_path, content)
        return content

    def read_file_str(self, storage_path: str, cache_store: bool = True) -> str:
        """Read the file from the source filesystem or from the source S3 bucket"""
        file_content = self.read_file_bytes(storage_path, cache_store=cache_store)
        return file_content.decode()

    def read_file_pickle(self, storage_path: str):
        """Read the file from the source filesystem or from the source S3 bucket
        and parse it as a pickle."""
        file_content = self.read_file_bytes(storage_path)
        return pickle.loads(file_content)

    def resolve_path(self, storage_path: str) -> str:
        """Resolve the storage path to an absolute path in the local filesystem or S3 bucket.
        :param storage_path: relative (within storage root) or absolute path to the source file
        :return: absolute path to the file in the local filesystem or S3 bucket
        """
        scheme, _, _ = self.parse_path(storage_path)
        if not os.path.isabs(storage_path) and not scheme:
            # Convert relative path to absolute
            return os.path.join(self.storage_root, storage_path)
        return storage_path

    def store_file_path(self, source_abs_path: str, storage_rel_path: str, overwrite: bool = True) -> str:
        """Store the file in the local filesystem or in the S3 bucket
        :param source_abs_path: abs path (or s3 URI) to the source file
        :param storage_rel_path: path to the storage file
        :param overwrite: if True, overwrite the file if it exists

        :return: relative path to the stored file
        """

        if not overwrite and self.file_exists(storage_rel_path):
            return storage_rel_path

        storage_abs_path = self.resolve_path(storage_rel_path)
        source_scheme, source_bucket, source_path = self.parse_path(source_abs_path)
        storage_scheme, storage_bucket, storage_path = self.parse_path(storage_abs_path)
        if source_scheme == "s3":
            if self.verbose:
                print(f"Downloading {source_abs_path} to {storage_abs_path}")
            if storage_scheme == "s3":
                self.aws.s3.copy_object(
                    CopySource=os.path.join(source_bucket, source_path), Bucket=storage_bucket, Key=storage_path
                )
                return storage_rel_path
            os.makedirs(os.path.dirname(storage_abs_path), exist_ok=True)
            try:
                self.aws.s3.download_file(Bucket=source_bucket, Key=source_path, Filename=storage_abs_path)
            except ClientError as e:
                print("S3 ERROR", e, e.response)
                if e.response["Error"]["Code"] in ("NoSuchKey", "404", 404):
                    raise FileNotFoundError(f"File not found: s3://{source_bucket}/{source_path}") from e
                raise
            return storage_rel_path
        else:
            if self.verbose:
                print(f"Downloading {source_abs_path} to {storage_abs_path}")
            os.makedirs(os.path.dirname(storage_abs_path), exist_ok=True)
            shutil.copy(source_abs_path, storage_abs_path)
            return storage_rel_path

    def store_file_bytes(self, file_bytes: bytes, storage_rel_path: str, overwrite: bool = True) -> str:
        """Store the file string in the source filesystem or source S3 bucket
        :param file_bytes: file bytes to store
        :param storage_rel_path: relative path to the storage file
        :param overwrite: if True, overwrite the file if it exists
        :returns: relative path to the stored file
        """

        if not overwrite and self.file_exists(storage_rel_path):
            return storage_rel_path

        storage_abs_path = self.resolve_path(storage_rel_path)
        storage_scheme, storage_bucket, storage_path = self.parse_path(storage_abs_path)
        if storage_scheme == "s3":
            if self.verbose:
                print(f"Uploading content to {storage_abs_path}")
            self.aws.s3.put_object(
                Body=file_bytes,
                Bucket=storage_bucket,
                Key=storage_path,
            )
        else:
            if self.verbose:
                print(f"Storing content to {storage_abs_path}")
            os.makedirs(os.path.dirname(storage_abs_path), exist_ok=True)
            with open(storage_abs_path, "wb") as f:
                f.write(file_bytes)

        return storage_rel_path

    def store_file_str(self, file_str: str, storage_rel_path: str, overwrite: bool = True) -> str:
        """Store the file string in the source filesystem or source S3 bucket
        :param file_str: file string to store
        :param storage_rel_path: relative path to the storage file
        :param overwrite: if True, overwrite the file if it exists
        """
        return self.store_file_bytes(file_str.encode(), storage_rel_path, overwrite=overwrite)

    def store_input(
        self,
        project_id: str,
        file_path: str = None,
        file_bytes: bytes = None,
        file_str: str = None,
        filename: str = None,
    ) -> str:
        """Store the file string in the source filesystem or source S3 bucket

        :param project_id: project ID
        :param file_path: path to the source file (when loading from path)
        :param file_bytes: file bytes to store (when loading from bytes)
        :param file_str: file string to store (when loading from string)
        :param filename: filename to use when storing the file (required when loading from bytes or string)
        """
        if file_path:
            assert file_bytes is None and file_str is None, "Provide only one of file_path, file_bytes, or file_str"
            if filename is None:
                filename = os.path.basename(file_path)
            # consider the input path to be relative to current dir, not the storage dir (useful in Jupyter)
            file_path = os.path.abspath(file_path)
            file_bytes = self.read_file_bytes(file_path)
        elif file_bytes is not None:
            assert file_str is None, "Provide only one of file_bytes or file_str"
            assert filename is not None, "filename must be provided when loading from bytes"
        elif file_str is not None:
            assert filename is not None, "filename must be provided when loading from string"
            file_bytes = file_str.encode()
        hash_str = get_hashed_path_for_bytes(file_bytes)
        storage_rel_path = os.path.join("project", project_id, "inputs", hash_str, filename)
        return self.store_file_bytes(file_bytes, storage_rel_path)

    def create_zip(self, storage_paths_by_dir: dict[str, list[str]]) -> bytes:
        """Create zip file
        :param storage_paths_by_dir: dictionary with the storage paths by directory
        :return: zip content with the stored files
        """

        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, "w") as zip_file:
            with ThreadPoolExecutor() as executor:
                # Fetch all files and collect their content
                futures = {
                    (file_path, executor.submit(self.read_file_bytes, file_path, cache_store=False)): dir_model
                    for dir_model, file_paths in storage_paths_by_dir.items()
                    for file_path in file_paths
                }

                # Process each future to write to ZIP file
                for (file_path, future), subdir in futures.items():
                    filename = os.path.basename(file_path)
                    file_data = future.result()  # this will raise exception if any
                    # Write the file to the zip with the desired structure
                    if file_data and filename:
                        zip_file.writestr(f"{subdir}/{filename}", file_data)
        zip_buffer.seek(0)
        return zip_buffer.read()

    def sync_file(self, storage_path: str, local_destination_path: str, link: bool = False):
        """Download/copy stored file to local directory
        :param storage_path: path to the stored file
        :param local_destination_path: path to the destination file
        :param link: if True, and if supported, create a symbolic link to the file instead of copying it
        """
        storage_path = self.resolve_path(storage_path)
        scheme, _, _ = self.parse_path(storage_path)
        os.makedirs(os.path.dirname(local_destination_path), exist_ok=True)
        if not scheme:
            # local to local copy
            if link:
                os.symlink(storage_path, local_destination_path)
            else:
                shutil.copy(storage_path, local_destination_path)
            return

        try:
            content = self.read_file_bytes(storage_path, cache_store=False)
        except IsADirectoryError:
            raise ValueError(f"Use sync_directory to sync a directory, got: {storage_path}")
        with open(local_destination_path, "wb") as f:
            f.write(content)

    def sync_files(self, storage_paths: list[str], local_destination_dir: str, preserve_subdirs=False):
        """Download/copy list of stored files to local directory
        :param storage_paths: list of files to be downloaded to destination directory
        :param local_destination_dir: path to destination directory (will be created if not exists)
        :param preserve_subdirs: if True (or if recursive=True), preserve subdirectories in the local destination directory
        """
        if isinstance(storage_paths, str):
            storage_paths = [storage_paths]

        if preserve_subdirs:
            for path in storage_paths:
                assert not os.path.isabs(path) and "://" not in path, (
                    f"Expected relative paths when preserve_subdirs=True, got: {path}"
                )

        os.makedirs(local_destination_dir, exist_ok=True)
        with ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(
                    self.sync_file,
                    file_path,
                    os.path.join(local_destination_dir, file_path if preserve_subdirs else os.path.basename(file_path)),
                )
                for file_path in storage_paths
            ]
            for future in futures:
                future.result()

    def is_local_path(self, path):
        scheme, netloc, relative_path = self.parse_path(path)
        return not scheme

    def sync_directory(self, source_dir: str, local_destination_dir: str, link=False):
        """Download/copy directory to local directory
        :param source_dir: source directory, path should be absolute or relative to storage root
        :param local_destination_dir: path to destination directory (will be created if not exists)
        :param link: Use symbolic links instead of copying files when possible.
        """
        source_dir = self.resolve_path(source_dir)
        if link and self.is_local_path(source_dir):
            os.symlink(source_dir, local_destination_dir)
            return

        os.makedirs(local_destination_dir, exist_ok=True)
        with ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(
                    self.sync_file, os.path.join(source_dir, file_path), os.path.join(local_destination_dir, file_path)
                )
                for file_path in self.list_dir(source_dir, recursive=True)
            ]
            for future in futures:
                future.result()

    def prepare_workflow_input(
        self, storage_path: str, workdir: str, input_bytes: bytes = None, name: str = None, custom_hash: str = None
    ) -> str:
        """Prepare workflow input files for submission by storing them in the provided scheduler workdir
        under unique hashes based on each file's contents.

        :param storage_path: path to the stored file
        :param workdir: destination (working directory of the scheduler)
        :param input_bytes: optional - do not read the storage_path, instead use the provided bytes
        :param name: optional - rename the file to provided name (without extension, original extension will be appended)
        :param custom_hash: use custom hash for the subdirectory name instead of using a generated hash of file contents
        """

        # TODO implement logic to avoid copying file if it's already accessible to the scheduler
        #  Two caveats:
        #  - The default scheduler.workdir points to a subdirectory different than the storage directory, for example ovo/workdir vs ovo/storage,
        #    so something like if storage_path.startswith(workdir) would not work
        #  - If name param is passed, we need to rename the file anyway - maybe a symlink would need to be created in that case

        if input_bytes is None:
            input_bytes = self.read_file_bytes(storage_path)

        if custom_hash is None:
            custom_hash = get_hashed_path_for_bytes(input_bytes)

        workdir_scheme, workdir_bucket, workdir_prefix = self.parse_path(workdir)

        if name is None:
            filename = os.path.basename(storage_path)
        else:
            filename = name + os.path.splitext(storage_path)[-1]

        if workdir_scheme == "s3":
            input_path = os.path.join("inputs", custom_hash, filename)
            input_prefix = os.path.join(workdir_prefix, input_path)
            s3_input_path = os.path.join(workdir, input_path)

            if self.file_exists(s3_input_path):
                return s3_input_path

            if self.verbose:
                print(f"Uploading {storage_path} to {s3_input_path}")

            self.aws.s3.put_object(Body=input_bytes, Bucket=workdir_bucket, Key=input_prefix)
            return s3_input_path
        else:
            input_dir = os.path.join(workdir, "inputs", custom_hash)
            os.makedirs(input_dir, exist_ok=True)
            input_path = os.path.join(input_dir, filename)

            if self.file_exists(input_path):
                return input_path

            with open(input_path, "wb") as f:
                f.write(input_bytes)
            return input_path

    def prepare_workflow_inputs(
        self,
        storage_paths: list[str],
        workdir: str,
        names: List[str] = None,
        return_paths=False,
        single_directory=False,
    ) -> str | list[str]:
        """
        Prepare workflow input files for submission by storing them in the provided scheduler workdir
        under unique hashes based on each file's contents.

        Each file will be stored in a separate subdirectory based on a hash of the file contents.
        This is useful when the same file can be submitted multiple times in different sets of files but should only be stored once.

        When return_paths=False (default), returns a txt file with the final file paths (or directly the file path in case of single input file).
        When return_paths=True, returns list of file paths.
        When single_directory=True, creates single directory and returns its path. Requires filenames to be unique.

        :param storage_paths: list of files to be downloaded to destination directory
        :param workdir: destination (working directory of the scheduler)
        :param names: rename each file to provided name (list of names of same length as storage_paths, without file extension, original extension will be added)
        """

        if single_directory:
            # Name subdirectory based on hash of paths to input files (not their contents)
            # This assumes that once files have been stored in storage, their content does not change
            custom_hash = get_hashed_path_for_bytes("_".join(storage_paths).encode())

            if names is not None:
                assert len(set(names)) == len(names), f"Names must be unique when single_directory=True, got: {names}"
            else:
                basenames = [os.path.basename(p) for p in storage_paths]
                assert len(set(basenames)) == len(basenames), (
                    f"Filenames must be unique when single_directory=True, got: {basenames}"
                )
        else:
            custom_hash = None

        if names is None:
            names = [None] * len(storage_paths)
        else:
            if len(names) != len(storage_paths):
                raise ValueError(
                    f"Mismatch between number of names ({len(names)}) and number of paths ({len(storage_paths)}). Ensure that each path has a corresponding name."
                )

        paths = [
            self.prepare_workflow_input(storage_path, workdir=workdir, name=name, custom_hash=custom_hash)
            for storage_path, name in zip(storage_paths, names)
        ]
        if single_directory:
            if not paths:
                raise ValueError("No paths to be prepared")
            subdirs = [path.rsplit("/", 1)[0] for path in paths]
            assert len(set(subdirs)) == 1, f"Expected all files to be in a single directory, got: {paths}"
            return subdirs[0] + "/"
        elif return_paths:
            return paths
        elif len(paths) == 1:
            # single input, return the file path directly without wrapping into a txt file
            return paths[0]
        else:
            # prepare txt file with one path per line
            return self.prepare_workflow_input(
                storage_path="input_pdb_paths.txt", workdir=workdir, input_bytes="\n".join(paths).encode("utf-8")
            )



from ovo.core.scheduler.base_scheduler import Scheduler, SchedulerTypes
from ovo.core.scheduler.healthomics_scheduler import HealthOmicsScheduler
from ovo.core.scheduler.nextflow_scheduler import NextflowScheduler



import streamlit as st
import json
import collections.abc


def load_json_from_file(filepath):
    try:
        with open(filepath, "r") as f:
            return json.load(f)
    except FileNotFoundError:
        st.error(f"Error: The file '{filepath}' was not found.")
        return None
    except json.JSONDecodeError:
        st.error(f"Error: The file '{filepath}' is not a valid JSON file.")
        return None
    except Exception as e:
        st.error(f"An unexpected error occurred: {e}")
        return None


def get_dict_diff(default: dict, custom: dict):
    diff = {}
    for key, custom_value in custom.items():
        if key not in default:
            diff[key] = custom_value
            continue

        default_value = default[key]
        if isinstance(custom_value, dict) and isinstance(default_value, dict):
            nested_diff = get_dict_diff(default_value, custom_value)
            if nested_diff:
                diff[key] = nested_diff
        elif custom_value != default_value:
            diff[key] = custom_value
    return diff


def merge_dictionaries(default: dict, diff: dict):
    if diff is None:
        return default.copy()

    merged = default.copy()
    for key, diff_value in diff.items():
        if key in merged and isinstance(merged[key], dict) and isinstance(diff_value, collections.abc.Mapping):
            merged[key] = merge_dictionaries(merged[key], diff_value)
        else:
            merged[key] = diff_value
    return merged



import pandas as pd
from Bio.Data.PDBData import protein_letters_3to1
from Bio.PDB import Structure
from scipy.spatial import cKDTree


def get_sequence(structure: Structure) -> str:
    """Get sequence of heavy (and light chain) in one string.

    Assumes only aa residues are in the structure and it contains coords for all the residues,
    e.g. predicted structures.

    First chain is asserted to be the heavy."""
    # FIXME
    chains = list(structure.get_chains())
    # assert 1 <= len(chains) <= 2
    # assert chains[0].id == 'H'  # first chain is Heavy

    # X for ambiguous/unknown (as in pep-patch)
    return "".join([protein_letters_3to1.get(r.resname, "X") for r in structure.get_residues()])


def get_atom_buriedness(structure, radii: list[float]):
    """Calculates the number of atoms within a specified radius for each atom in a PDB file.

    Parameters:
    structure (Structure): Biopython structure.
    radii (List[float]): Radii to calculate buriedness.

    Returns DataFrame with atom number, residue_id (str, possibly with insertion code),
        chain_id, and buriedness (atom count in query atom's vicinity) for each radius.
    """
    # Extract atom coordinates and info
    atoms = [
        (atom, atom.get_coord())
        for chain in structure.get_chains()
        for residue in chain.get_residues()
        for atom in residue.get_atoms()
    ]

    # Create a KDTree
    kdtree = cKDTree([coord for _, coord in atoms])

    # Prepare a dictionary to hold the results
    results = {}

    # Calculate buriedness for each atom
    for atom, coord in atoms:
        _, res_number, insertion_code = atom.get_parent().get_id()
        res_id = str(res_number) if insertion_code in (" ", "") else str(res_number) + insertion_code  # '80' or '80A'

        atom_key = (atom.get_serial_number(), res_id, atom.get_parent().get_parent().get_id())
        results[atom_key] = {}
        for radius in radii:
            # Count atoms within the radius
            count = kdtree.query_ball_point(coord, radius, return_length=True) - 1  # exclude the atom itself
            results[atom_key][f"buriedness_{radius}"] = count

    # Create a DataFrame from the results
    df = pd.DataFrame.from_dict(results, orient="index").reset_index()
    df.columns = ["atom_number", "residue_id", "chain_id"] + [f"buriedness_{r}" for r in radii]

    # Create a residue_serial column that increments for each unique residue_id
    df = df.sort_values(
        "atom_number"
    ).reset_index()  # first sort to get the same order as in PDB file (if biopython atom order is not guaranteed)
    df["residue_serial"] = df.groupby(
        ["residue_id", "chain_id"], sort=False
    ).ngroup()  # group by also chain (sort=False, so should be safe without chain gb even)

    return df



import pandas as pd
import requests
import re
import numpy as np
import typing
import os

from io import BytesIO

from ovo.app.components.molstar_custom_component.color_picker import ColorPicker, mix_colors
from ovo.app.components.molstar_custom_component.dataclasses import (
    ContigSegment,
)


class ContigsParser:
    """Class for parsing the contigs definition."""

    def get_chain_and_indices_fixed_region(self, region: str):
        """Get the chain and region residue indices.
        e.g.: From A123-131 return (A, 123, 131)."""
        chain_inp = "".join(re.findall(r"[a-zA-Z]+", region[0]))
        first_idx_inp = "".join(re.findall(r"[0-9]+", region[0]))
        if len(region) > 1:
            end_idx_inp = "".join(re.findall(r"[0-9]+", region[1]))
        else:
            end_idx_inp = first_idx_inp

        assert len(chain_inp) <= 2, f"Chains must have a maximum of 2 characters. {region}"
        assert first_idx_inp != "", f"Expected to find a number value for the contig {region}."
        assert int(first_idx_inp) <= int(end_idx_inp), f"Expected {first_idx_inp} <= {end_idx_inp} for {region}."
        assert len(region) < 3, f"Expected just 1-2 numbers in {region}"

        return chain_inp, first_idx_inp, end_idx_inp

    def get_output_chain(self, has_generated_before_chain_break: bool):
        """Get the output chain. Currently, we assume that anything with a "generated region" until there is a chain break
        results in chain A, otherwise returns chain B."""
        return "A" if has_generated_before_chain_break else "B"

    def parse_region_generated(self, region: str, last_parsed_res: int, has_generated_before_chain_break: bool):
        """Parse regions that are defined to be generated by RFdiffusion."""
        split_region = region.split("-")
        out_res_start = last_parsed_res + 1

        if len(split_region) == 1:
            split_region.append(split_region[0])

        assert len(split_region) <= 2 and len(split_region) > 0, f"Expected one or two coordinates, got: {region}"
        # we do not want the case of 10-20
        assert split_region[0] == split_region[1], f"Expected fixed region length, got: {region}"

        # without postprocessing it is not possible to guess the final boundaries
        out_res_end = out_res_start + int(split_region[0]) - 1

        return ContigSegment(
            value=region,
            length=int(split_region[0]),
            type="generated",
            input_res_start=-1,  # we do not care about input here as it does not exist
            input_res_end=-1,
            input_res_chain="?",
            out_res_start=last_parsed_res + 1,
            out_res_end=out_res_end,
            out_res_chain=self.get_output_chain(has_generated_before_chain_break),
        )

    def parse_region_fixed_no_mapping(self, region: str):
        """Parse fixed regions with an identity mapping."""
        split_region = region.split("-")
        first_chain_inp, first_idx_inp, end_idx_inp = self.get_chain_and_indices_fixed_region(split_region)

        return ContigSegment(
            value=region,
            type="fixed",
            length=int(end_idx_inp) - int(first_idx_inp) + 1,
            input_res_start=int(first_idx_inp),
            input_res_end=int(end_idx_inp),
            input_res_chain=first_chain_inp,  # should be OK as long as we assume there are no things such as A100-B150
            out_res_start=int(first_idx_inp),
            out_res_end=int(end_idx_inp),
            out_res_chain=first_chain_inp,  # should be OK as long as chain breaks are used properly
        )

    def parse_region_fixed(
        self,
        region: str,
        ref_indices: list[list[typing.Union[str, np.int64, bool]]],
        new_indices: list[list[typing.Union[str, np.int64, bool]]],
        has_generated_before_chain_break: bool,
    ):
        """Parse fixed regions with corresponding mapping defined in the .trb file."""
        split_region = region.split("-")
        first_chain_inp, first_idx_inp, end_idx_inp = self.get_chain_and_indices_fixed_region(split_region)

        # now let's find the corresponding values in the reference/new indices list (otherwise return -1)
        ref_indices_first_index = next(
            (
                i
                for i, item in enumerate(ref_indices)
                if item[0] == first_chain_inp and item[1] == np.int64(first_idx_inp) and item[2] == False
            ),
            -1,
        )

        ref_indices_last_index = next(
            (
                i
                for i, item in enumerate(ref_indices)
                if item[0] == first_chain_inp and item[1] == np.int64(end_idx_inp) and item[2] == False
            ),
            -1,
        )

        assert ref_indices_first_index > -1, f"Could not find mapping for residue {first_chain_inp}{first_idx_inp}"
        assert ref_indices_last_index > -1, f"Could not find mapping for residue {first_chain_inp}{end_idx_inp}"

        # Set the flag to already used.
        new_indices[ref_indices_first_index][2] = True
        new_indices[ref_indices_last_index][2] = True

        ref_indices[ref_indices_first_index][2] = True
        ref_indices[ref_indices_last_index][2] = True

        # now we have the indices so we might go on and use them to find the mapping
        # the mapping will use the same indices
        first_res_mapping = new_indices[ref_indices_first_index]
        last_res_mapping = new_indices[ref_indices_last_index]

        return ContigSegment(
            value=region,
            type="fixed",
            length=int(end_idx_inp) - int(first_idx_inp) + 1,
            input_res_start=int(first_idx_inp),
            input_res_end=int(end_idx_inp),
            input_res_chain=first_chain_inp,  # should be OK as long as we assume there are no things such as A100-B150
            out_res_start=int(first_res_mapping[1]),
            out_res_end=int(last_res_mapping[1]),
            out_res_chain=self.get_output_chain(
                has_generated_before_chain_break
            ),  # should be OK as long as chain breaks are used properly
        )

    def parse_contigs_str(self, contigs_string: str = "", include_generated=False) -> list[ContigSegment]:
        return self._parse_contigs_with_tuples(contigs=contigs_string, include_generated=include_generated)

    def parse_contigs_trb(self, contigs_trb: str | dict = "") -> list[ContigSegment]:
        """Parse the contigs from a given .trb file or .trb parsed dictionary.
        Can be either an http(s) URL, a system path to the .trb file or parsed .trb file content directly (checked in this order).
        """
        # notice that in this case we are using the contigs from the .trb file, not the original ones that were defined by the user
        # first, we check if "trb" is an URL, then we check for a file and lastly we treat "trb" as the unpacked pickle
        if "http://" in contigs_trb or "https://" in contigs_trb:
            r = requests.get(contigs_trb)
            trb_file = pd.read_pickle(BytesIO(r.content))
        elif (isinstance(contigs_trb, str) or isinstance(contigs_trb, os.PathLike)) and os.path.exists(contigs_trb):
            trb_file = pd.read_pickle(contigs_trb)
        elif isinstance(contigs_trb, dict):
            trb_file = contigs_trb
        else:
            raise RuntimeError(
                f"Provided an invalid .trb definition. Cannot parse: {contigs_trb}, {type(contigs_trb)}."
            )

        # contigs: str = trb_file["config"]["contigmap"]["contigs"][0].replace(" ", "/0 ")
        contigsList: list[str] = trb_file["sampled_mask"]
        newContigs = []

        for idx, contig in enumerate(contigsList):
            contig = contig.removesuffix("/0")
            newContigs.append(contig)

        contigs = "/0 ".join(newContigs)

        if "complex_con_ref_pdb_idx" in trb_file:
            ref_idx = trb_file["complex_con_ref_pdb_idx"]
        else:
            assert "con_ref_pdb_idx" in trb_file, "Could not find con_ref_pdb_idx in the .trb file"
            ref_idx = trb_file["con_ref_pdb_idx"]

        if "complex_con_hal_pdb_idx" in trb_file:
            hal_idx = trb_file["complex_con_hal_pdb_idx"]
        else:
            assert "con_hal_pdb_idx" in trb_file, "Could not find con_hal_pdb_idx in the .trb file"
            hal_idx = trb_file["con_hal_pdb_idx"]

        return self.parse_contigs_ref(contigs, ref_idx=ref_idx, hal_idx=hal_idx)

    def parse_contigs_ref(self, contigs: str, ref_idx: list[tuple], hal_idx: list[tuple]):
        """Parse the contigs and return segments with output numbering."""
        reference_tuples = []
        for tuple in ref_idx:
            reference_tuples.append([tuple[0], tuple[1], False])

        mapped_tuples = []
        for tuple in hal_idx:
            mapped_tuples.append([tuple[0], tuple[1], False])

        return self._parse_contigs_with_tuples(
            contigs=contigs, mapped_tuples=mapped_tuples, reference_tuples=reference_tuples, include_generated=True
        )

    def _parse_contigs_with_tuples(
        self, contigs: str, mapped_tuples: list = [], reference_tuples: list = [], include_generated=True
    ) -> list[ContigSegment]:
        """
        Parse the contigs from the .trb file. If no .trb file path is defined, then the parser treats fixed regions with no mapping.

        Parameters
        ----------
            contigs: str
                Contigs string definition (e.g.: A30-54/10/A70-75).
            mapped_tuples: list
                Mapped residues from the .trb contigs definition.
            reference_tuples: list
                Mapped reference residues from the .trb contigs definition.

        Returns
        -------
            Parsed contig regions
        """
        if not contigs:
            return []

        parsed_segments: list[ContigSegment] = []
        color_picker = ColorPicker()

        last_parsed_residue = 0
        has_generated_before_chain_break = False

        # Here, we check if there is a generated segment in each "bigger region" separated by a chain break.
        # For example, for these regions    ['A57-67/10-10/A34-42/10-10/A90-102/10-10/A126-137', 'H1-50', 'L1-114', 'A50-67/20-20/A70-77', 'H60-80']
        # we expect these results:          [True, False, False, True, False]
        split_regions_by_chainbreak = contigs.split("/0 ")
        regions_contain_generated = []
        for split_reg in split_regions_by_chainbreak:
            found = False

            for seg in split_reg.split("/"):
                seg = seg.replace("0 ", "").replace(" ", "")

                if not any(c.isalpha() for c in seg):  # this means there is no letter -> a generated specification
                    found = True
                    break

            regions_contain_generated.append(found)

        chainbreak_idx = 0
        has_generated_before_chain_break = regions_contain_generated[chainbreak_idx]

        split_segments = contigs.split("/")
        for segment in split_segments:
            if segment.startswith("0 "):
                # For each of the bigger regions, we have to check if there was a generated segment inside to
                chainbreak_idx += 1
                has_generated_before_chain_break = regions_contain_generated[chainbreak_idx]

                last_parsed_residue = 0

                new_segment_split = segment.split(" ")
                assert len(new_segment_split) == 2, f"Expected a format of '/0 <chain>..-..', got {segment}"

                segment = new_segment_split[1]

            if not any(c.isalpha() for c in segment):  # this means there is no letter -> a generated specification
                if not include_generated:
                    # for contigs defined in a string, we do not parse the generated regions
                    continue
                parsed_segment = self.parse_region_generated(
                    segment, last_parsed_residue, has_generated_before_chain_break
                )
            else:
                if len(reference_tuples) == 0:
                    # there is no mapping in this case
                    parsed_segment = self.parse_region_fixed_no_mapping(segment)
                else:
                    parsed_segment = self.parse_region_fixed(
                        segment, reference_tuples, mapped_tuples, has_generated_before_chain_break
                    )

            last_parsed_residue = parsed_segment.out_res_end
            parsed_segments.append(parsed_segment)

        # assign colors on segments sorted in original order, to get a consistent coloring independent of reordering
        sorted_segments = sorted(parsed_segments, key=lambda x: (x.input_res_chain, x.input_res_start, x.out_res_start))
        for seg in sorted_segments:
            if seg.type == "generated":
                seg.color = "#44aa44"  # fallback color - will be changed below
            else:
                seg.color = color_picker((seg.input_res_start, seg.input_res_end, seg.input_res_chain))

        # assign color to generated segments based on neighboring fixed segments
        for prev_seg, seg, next_seg in zip(
            [None] + parsed_segments[:-1], parsed_segments, parsed_segments[1:] + [None]
        ):
            if seg.type == "generated":
                if prev_seg:
                    # mid generated segment
                    seg.color = mix_colors(mix_colors(prev_seg.color, "#ffffff"), "#ffffff")
                elif next_seg:
                    # start segment (N term)
                    seg.color = "#eeeeee"

        return parsed_segments



import json
import os

import streamlit as st

from ovo import config, db, schedulers
from ovo.core.database.models_rfdiffusion import RFdiffusionWorkflow, ProteinMPNNParams
from ovo.app.components.acceptance_thresholds_components import thresholds_input_component
from ovo.app.components.navigation import open_first_section
from ovo.app.utils.bindcraft_utils import load_json_from_file, get_dict_diff, merge_dictionaries
from ovo.app.utils.cached_db import get_cached_round
from ovo.app.utils.testing import is_test_dialog_shown
from ovo.core.auth import get_username
from ovo.core.database import (
    descriptors_rfdiffusion,
)
from ovo.core.database.descriptors_refolding import REFOLDING_TESTS_BY_TYPE
from ovo.core.database.models_rfdiffusion import RFdiffusionBinderDesignWorkflow, RFdiffusionScaffoldDesignWorkflow
from ovo.core.database.models import Pool, Round, Workflow
from ovo.core.logic.design_logic import submit_design_workflow
from ovo.core.logic.round_logic import get_or_create_project_rounds


def pool_submission_inputs(page_key: str):
    rounds_by_id = get_or_create_project_rounds(project_id=st.session_state.project.id)
    round_ids = list(rounds_by_id.keys())

    if st.session_state.pool_inputs.get(page_key):
        round_id, pool_name, pool_description = st.session_state.pool_inputs[page_key]
        if round_id is None:
            round_id = round_ids[-1]
    else:
        round_id, pool_name, pool_description = round_ids[-1], "", ""

    if st.session_state.get("new_round_id"):
        # When a new round is created, select it by default
        round_id = st.session_state["new_round_id"]
        del st.session_state["new_round_id"]

    left, right, _ = st.columns([1, 1, 2], vertical_alignment="bottom")
    with left:
        round_id = st.selectbox(
            "Project Round",
            format_func=lambda i: rounds_by_id[i].name,
            # TODO force change does the trick in streamlit 1.50.0
            #  because the component gets recreated when the options change
            #  However this might change in the future.
            #  We should find another way to force the selectbox to update when a new round is created.
            #  This is covered in unit tests.
            options=round_ids,
            index=round_ids.index(round_id) if round_id in round_ids else len(round_ids) - 1,
            key="selected_round_id",
        )

    with right:
        if st.button("Create new Round") or is_test_dialog_shown("create_round"):
            create_new_round_dialog()

    with st.columns([2, 1])[0]:
        pool_name = st.text_input(
            "Pool name",
            placeholder="Name your submission",
            value=pool_name,
            key="input_pool_name",
        )
        previous_pools = db.select(Pool, round_id=round_id, limit=3, order_by="-created_date_utc")
        if previous_pools:
            st.write(f"*Previous pool names: {', '.join([p.name for p in previous_pools if p.name])}*")
        if db.count(Pool, round_id=round_id, name=pool_name):
            st.error(f"Pool '{pool_name}' already exists in this round, please choose a different name.")

    pool_description = st.text_area(
        "Pool description (optional)",
        placeholder="Describe your submission here",
        key="input_pool_description",
        value=pool_description,
    )

    st.session_state.pool_inputs[page_key] = round_id, pool_name, pool_description


def get_pool_inputs(page_key: str) -> tuple[str, str, str]:
    if not st.session_state.pool_inputs.get(page_key):
        # This should never happen since pool_submission_inputs() should be called before
        st.error("Pool inputs not found.")
        st.stop()

    round_id, pool_name, pool_description = st.session_state.pool_inputs[page_key]
    return round_id, pool_name, pool_description


def review_workflow_submission(page_key: str):
    """Show a summary of the workflow settings and submission modal button."""
    workflow = st.session_state.workflows[page_key]

    round_id, pool_name, pool_description = get_pool_inputs(page_key)

    if not pool_name:
        st.error("Please provide a pool name in the submission tab.")
        return

    # Validate workflow parameters
    try:
        workflow.validate()
    except ValueError as e:
        st.write("Errors found in workflow parameters:")
        st.error(e)
        return

    project_round = get_cached_round(round_id)

    # Pool name and description
    st.write(f"**Project:** {st.session_state.project.name}")
    st.write(f"**Round:** {project_round.name}")
    st.write(f"**Pool name:** {pool_name}")
    if db.count(Pool, round_id=round_id, name=pool_name):
        st.error(f"Pool '{pool_name}' already exists in this round, please choose a different name.")
    if pool_description:
        st.markdown(f"**Pool description**: {pool_description}")

    # Workflow parameters
    st.markdown(
        """
        <style>
        .stTable table {
            width: auto !important;
        }
        .stTable > div {
            display: inline-block;
        }
        </style>
        """,
        unsafe_allow_html=True,
    )
    st.table(
        workflow.get_table_row(name="value").apply(lambda v: ", ".join(map(str, v)) if isinstance(v, list) else v),
    )

    if st.button("Continue to submission :material/arrow_forward_ios:", key="continue") or is_test_dialog_shown(
        "submission_dialog"
    ):
        submit_workflow_dialog(
            page_key=page_key,
            workflow=workflow,
            round_id=round_id,
            pool_name=pool_name,
            pool_description=pool_description,
        )


@st.dialog("Submit workflow", width="large")
def submit_workflow_dialog(page_key: str, workflow: Workflow, round_id: str, pool_name: str, pool_description: str):
    # Create "empty" element to enable clearing the contents after submitting
    content = st.empty()
    with content.container():
        # Scheduler dropdown
        with st.columns(2)[0]:
            scheduler_key = st.selectbox(
                "Scheduler", options=schedulers.keys(), format_func=lambda k: schedulers[k].name
            )

        # Time estimate
        st.write(workflow.get_time_estimate(schedulers[scheduler_key]))

        # Submit button
        with st.columns([2, 1])[-1]:
            submitted = st.button("Submit", type="primary", width="stretch", key="submit_run")

    if submitted:
        # Clear dialog window contents
        content.empty()

        # Validate workflow parameters
        try:
            workflow.validate()
        except ValueError as e:
            st.error(e)
            return

        st.write("Submitting job... 🚀")
        submit_design_workflow(
            workflow,
            scheduler_key,
            round_id=round_id,
            pool_name=pool_name,
            pool_description=pool_description,
            return_existing=False,
        )

        st.session_state.flash_success = "Job submitted successfully"
        open_first_section(page_key)
        # Switch to jobs page and clear session state
        st.session_state.workflows.pop(page_key, None)
        st.session_state.pool_inputs.pop(page_key, None)
        st.switch_page("app/pages/jobs/jobs.py")


@st.dialog("Create new round of designs")
def create_new_round_dialog():
    project_id = st.session_state.project.id
    last_round = list(get_or_create_project_rounds(project_id=project_id).values())[-1]
    next_name = None
    if last_round.name.split()[-1].isnumeric():
        last_number = int(last_round.name.split()[-1])
        next_name = f"Round {last_number + 1}"
    round_name = st.text_input("Round name", value=next_name, key="round_name")

    if st.button("Create round", disabled=not round_name, key="create_round"):
        project_round = Round(project_id=project_id, name=round_name, author=get_username())
        db.save(project_round)
        st.session_state.new_round_id = project_round.id
        st.rerun()


def show_rfdiffusion_binder_seq_design_inputs(workflow: RFdiffusionWorkflow):
    seq_design_options = {
        "ligandmpnn": "LigandMPNN (ProteinMPNN weights)",
        "fastrelax": "ProteinMPNN-FastRelax",
    }
    captions = {
        "ligandmpnn": "Sequence design with ProteinMPNN model through its LigandMPNN re-implementation. "
        "Designs will not be relaxed as part of sequence design, only repacked with LigandMPNN OpenMM logic. "
        "Designs that pass refolding criteria will be relaxed during ddG scoring.",
        "fastrelax": "Cycles of ProteinMPNN design followed by Rosetta FastRelax relaxation "
        "based on Bennet et al. 2023 'Improving de novo protein binder design with deep learning'. "
        "Sequences from all cycles are kept.",
    }
    if workflow.protein_mpnn_params.fastrelax_cycles:
        seq_design_method = "fastrelax"
    else:
        seq_design_method = "ligandmpnn"
    options = list(seq_design_options)
    seq_design_method = st.radio(
        "Sequence design method",
        options=options,
        format_func=seq_design_options.get,
        index=options.index(seq_design_method) if seq_design_method in options else None,
        captions=[captions[o] for o in options],
        key="seq_design_method",
    )
    if seq_design_method == "fastrelax":
        if not config.props.pyrosetta_license:
            st.error(
                "Use of the PyRosetta license is disabled in this instance of OVO. "
                "If you want to use PyRosetta FastRelax for interface side-chains relaxation and "
                "you have a license (or in case of non-commercial use), "
                "enable props.pyrosetta_license in your config.yml"
            )
            st.stop()
        if workflow.protein_mpnn_params.fastrelax_cycles == 0:
            # Initialize at 3 cycles
            workflow.protein_mpnn_params.fastrelax_cycles = 3
        workflow.protein_mpnn_params.num_sequences = 1
        with st.columns([1, 2])[0]:
            workflow.protein_mpnn_params.fastrelax_cycles = st.number_input(
                "Number of FastRelax cycles (ProteinMPNN design followed by FastRelax relaxation)",
                min_value=1,
                max_value=5,
                value=workflow.protein_mpnn_params.fastrelax_cycles,
                key="fastrelax_cycles",
            )
        st.caption(
            "Note: Compared to Bennet et al. 2023, we skip the last ProteinMPNN design step "
            "so that all designs have undergone relaxation consistently. For example, with 2 cycles, "
            "the workflow actually performs (Design + Relax) * 2 instead of (Design + Relax) * 2 + Design."
        )
    else:
        if workflow.protein_mpnn_params.fastrelax_cycles:
            # Re-initialize to default number of sequences when switching back from FastRelax
            workflow.protein_mpnn_params.num_sequences = ProteinMPNNParams().num_sequences
            workflow.protein_mpnn_params.fastrelax_cycles = 0
        with st.columns([1, 2])[0]:
            workflow.protein_mpnn_params.num_sequences = st.number_input(
                f"Number of sequence designs per backbone",
                min_value=1,
                max_value=config.props.mpnn_sequences_limit,
                value=workflow.protein_mpnn_params.num_sequences,
                key="num_sequences",
            )


def show_rfdiffusion_advanced_settings(workflow: RFdiffusionWorkflow):
    #
    # FIXME remove this hack
    #
    if not config.props.pyrosetta_license:
        for key, threshold in workflow.acceptance_thresholds.items():
            if "rosetta" in key and threshold.enabled:
                st.warning(f"Disabling PyRosetta acceptance threshold as PyRosetta license is not available: {key}")
                threshold.enabled = False
    #
    #

    st.markdown("Advanced settings")

    with st.expander(label="Show advanced settings"):
        if get_username() in config.auth.admin_users:
            st.markdown("#### Admin settings")

            workflow.rfdiffusion_params.batch_size = st.number_input(
                "Batch size",
                min_value=1,
                max_value=200,
                value=workflow.rfdiffusion_params.batch_size,
                key="batch_size",
            )

        st.markdown("#### RFdiffusion")

        workflow.rfdiffusion_params.timesteps = st.number_input(
            "Num RFdiffusion timesteps (T)",
            min_value=1,
            value=workflow.rfdiffusion_params.timesteps,
            key="timesteps",
        )
        st.caption(
            ":material/info: Number of denoising diffusion steps determines the granularity of the diffusion process. "
            "Higher values lead to better quality structures but also longer runtime. "
        )

        if workflow.is_instance(RFdiffusionScaffoldDesignWorkflow):
            workflow.rfdiffusion_params.contigmap_length = (
                st.text_input(
                    "Sequence length limit (contigmap.length)",
                    placeholder="For example 123 or 123-456",
                    value=workflow.rfdiffusion_params.contigmap_length,
                    key="contigmap_length",
                )
                or None
            )
            st.caption(
                ":material/info: When using multiple generated segments with random length range (for example 10-100), "
                "make sure that randomly sampling those lengths will produce contigs within the total length range."
            )

            workflow.rfdiffusion_params.inpaint_seq = (
                st.text_input(
                    "Sequence inpainting regions (contigmap.inpaint_seq)",
                    placeholder="For example A12-34/A45-67/B89-90",
                    value=workflow.rfdiffusion_params.inpaint_seq,
                    key="inpaint_seq",
                )
                or None
            )
            st.caption(
                ":material/info: Input structure regions present in the contig but to be redesigned with ProteinMPNN. "
                "This argument tells RFdiffusion that the sequence will be redesigned "
                "so it can better pack against those regions."
            )

        workflow.rfdiffusion_params.backbone_filters = (
            st.text_input(
                "Hard filters for backbone designs",
                placeholder="For example pydssp_helix_percent<50,N_contact_hotspots>=5",
                value=workflow.rfdiffusion_params.backbone_filters,
                key="backbone_filters",
            )
            or None
        )
        st.caption(
            "This filter can be used to completely discard undesirable backbone designs "
            "and avoid wasting compute resources on their sequence design and structure prediction."
        )
        st.caption(f"Supported filtering metrics: {', '.join(descriptors_rfdiffusion.BACKBONE_METRIC_FIELD_NAMES)}")

        workflow.rfdiffusion_params.run_parameters = st.text_input(
            "Additional commandline arguments for RF diffusion",
            placeholder="For example contigmap.inpaint_str=[123-456] contigmap.inpaint_str_strand=[123-456]",
            value=workflow.rfdiffusion_params.run_parameters,
            key="rfdiff_run_parameters",
        )
        st.caption(":material/info: Reference: https://github.com/RosettaCommons/RFdiffusion.")
        if workflow.rfdiffusion_params.run_parameters:
            st.warning("Additional parameters can override previously selected parameters.")

        st.markdown("#### Protein MPNN ")

        workflow.protein_mpnn_params.sampling_temp = st.number_input(
            "Sampling temperature protein MPNN",
            format="%g",
            value=workflow.protein_mpnn_params.sampling_temp
            or (0.0001 if workflow.is_instance(RFdiffusionBinderDesignWorkflow) else 0.1),
            key="sampling_temp",
        )
        st.caption(
            ":material/info: ProteinMPNN is a deterministic neural network, outputting residue probabilities. "
            "The temperature parameter is used to select less probable residues during sequence sampling. "
            "Higher temperature leads to higher diversity, temperature closer to 0 leads to less diversity and higher ProteinMPNN score. "
            "Suggested values 0.0001, 0.1, 0.15, 0.2, 0.25, 0.3. "
        )

        workflow.protein_mpnn_params.omit_aa = st.text_input(
            "Omitted residue types from sequence design",
            value=workflow.protein_mpnn_params.omit_aa,
            key="omit_aa",
        )

        workflow.protein_mpnn_params.bias_aa = st.text_input(
            "Amino acid bias for ProteinMPNN",
            placeholder="For example F:1.5,K-1.2",
            value=workflow.protein_mpnn_params.bias_aa,
            key="mpnn_bias",
        )
        st.caption(
            ":material/info: Negative values reduce the probability of sampling given amino acid, positive values increase it, zero is neutral (default)."
        )

        st.markdown("#### Refolding tests")

        # For example 'scaffold' or 'binder'
        supported_refolding_tests = REFOLDING_TESTS_BY_TYPE[workflow.get_refolding_design_type()]
        test_options = list(supported_refolding_tests.keys())

        workflow.refolding_params.primary_test = st.radio(
            "Refolding test (Structure prediction method used to compare designed and predicted structure)",
            options=test_options,
            format_func=lambda k: f"**{supported_refolding_tests[k][0]}** {supported_refolding_tests[k][1]}",
            key="refolding_test",
            index=test_options.index(workflow.refolding_params.primary_test)
            if workflow.refolding_params.primary_test in test_options
            else None,
        )

        if workflow.refolding_params.primary_test == "esmfold":
            workflow.refolding_params.esmfold_fp16 = st.checkbox(
                "Use fp16 mode in ESMFold",
                help="Use half-precision mode in ESMFold to reduce memory consumption and speed up predictions. This may lead to slightly less accurate predictions.",
                value=workflow.refolding_params.esmfold_fp16,
                key="esmfold_fp16",
            )

        new_thresholds = thresholds_input_component(selected_thresholds=workflow.acceptance_thresholds)
        if new_thresholds != workflow.acceptance_thresholds:
            workflow.acceptance_thresholds = new_thresholds
            st.rerun()
            return


def customize_json_settings_component(diff: dict | None, default: dict, key_suffix: str):
    """
    Allows to modify default advanced an filter setting in a text field and returns only the differences between custom and default.

    Args:
        diff: A dictionary with previously saaved customizations. (Differences from the default)
        default: The base dictionary with default settings.
        key_suffix: Unique string used to crate keys for Streamlit components.

    Returns:
        Newly calculated differences and boolean indicating whether the component is in edit mode or not.
    """
    merged_settings = merge_dictionaries(default, diff)
    new_diff = get_dict_diff(default, merged_settings)
    merged_json = json.dumps(merged_settings, indent=4)

    if new_diff:
        st.warning(f"You have customized the '{key_suffix}' settings. See changes below:")
        st.json(new_diff, expanded=False)

    col1, info_col = st.columns([0.2, 0.8])
    with col1:
        edit_mode = st.toggle("Edit", key=f"edit_toggle_{key_suffix}")

    with st.columns([1.2, 1])[0]:
        if edit_mode:
            with info_col:
                st.info(
                    "You are in edit mode. Changes are processed live. Toggle 'Edit' off to see the changes.",
                    icon="✍️",
                )

            editor_key = f"editor_text_{key_suffix}"
            response_text = st.text_area(
                "Edit JSON settings",
                height=400,
                value=merged_json,
                key=editor_key,
                label_visibility="collapsed",
            )

            try:
                response_dict = json.loads(response_text)
                new_diff = get_dict_diff(default, response_dict)
                if not new_diff:
                    new_diff = None
            except json.JSONDecodeError:
                st.error("The text is not valid JSON. Please correct it.", icon="🚨")
                pass  # Keep the last valid diff

        else:
            st.code(merged_json, language="json", line_numbers=True, height=400)

        return new_diff, edit_mode


@st.fragment()
def show_bindcraft_advanced_settings(workflow: "BindCraftBinderDesignWorkflow"):
    advanced_settings_path, filter_settings_path = workflow.get_settings_paths()
    advanced_default_dict = load_json_from_file(advanced_settings_path)
    filter_default_dict = load_json_from_file(filter_settings_path)

    if advanced_default_dict is None or filter_default_dict is None:
        st.error("Cannot display settings component because the default settings failed to load.")
        return

    with st.expander(label="Advanced settings and filter settings"):
        st.markdown("#### Advanced Settings")
        advanced_key = os.path.basename(advanced_settings_path).removesuffix(".json")

        st.write(
            f"""
            Settings reference: [github.com/martinpacesa/BindCraft](https://github.com/martinpacesa/BindCraft?tab=readme-ov-file#advanced-settings)
            
            The advanced settings json file **`{advanced_key}`** 
            was selected based on the design protocol choices above. 
            You can further customize the settings using the editor below.
            """
        )

        custom_advanced, advanced_edit_mode = customize_json_settings_component(
            diff=workflow.bindcraft_params.custom_advanced_settings,
            default=advanced_default_dict,
            key_suffix=advanced_key,
        )
        workflow.bindcraft_params.custom_advanced_settings = custom_advanced

        if workflow.bindcraft_params.custom_advanced_settings and not advanced_edit_mode:
            if st.button("Reset Advanced Settings to Default", key="reset_advanced", type="primary"):
                workflow.bindcraft_params.custom_advanced_settings = None
                # editor_key = f"editor_text_{advanced_key}"
                # if editor_key in st.session_state:
                #     del st.session_state[editor_key]
                st.rerun()

        st.divider()

        st.markdown("#### Filter Settings")

        filter_key = os.path.basename(filter_settings_path).removesuffix(".json")

        st.write(
            f"""
            Settings reference: [github.com/martinpacesa/BindCraft](https://github.com/martinpacesa/BindCraft?tab=readme-ov-file#advanced-settings)
            
            The filter settings json file **`{filter_key}`** 
            was selected based on the design protocol choices above. 
            You can further customize the settings using the editor below.
            """
        )

        custom_filter, filter_edit_mode = customize_json_settings_component(
            diff=workflow.bindcraft_params.custom_filter_settings, default=filter_default_dict, key_suffix=filter_key
        )
        workflow.bindcraft_params.custom_filter_settings = custom_filter

        if workflow.bindcraft_params.custom_filter_settings and not filter_edit_mode:
            if st.button("Reset Filter Settings to Default", key="reset_filter", type="primary"):
                workflow.bindcraft_params.custom_filter_settings = None
                # editor_key = f"editor_text_{filter_key}"
                # if editor_key in st.session_state:
                #     del st.session_state[editor_key]
                st.rerun()



import os
import time
import traceback
from abc import ABC
from multiprocessing.managers import BaseManager
from queue import Queue
from typing import Any


class SimpleQueueMixin(ABC):
    """Class that adds a simple queue-based worker scheduler to a submission system."""

    def has_queue(self):
        return "queue" in self.submission_args

    def _queue_get_address(self):
        host_port = self.submission_args.get("queue", None)
        if not host_port:
            raise ValueError(
                f"No host:port specified for this scheduler, 'queue' field missing in submission_args of scheduler {self.name}"
            )
        if ":" not in host_port:
            raise ValueError(f"Invalid host:port {host_port} specified for scheduler {self.name}")
        host, port_str = host_port.split(":", 1)
        try:
            port = int(port_str)
        except ValueError:
            raise ValueError(f"Invalid port {port_str} specified for scheduler {self.name}")
        return host, port

    def queue_put(self, job_id: str, task: Any):
        """Put a task into the queue to be processed by a worker."""
        queue = self.connect_to_queue()
        queue.put((job_id, task))

    def queue_size(self, job_id: str = None) -> int | None:
        """Get number of tasks in the queue (not running yet)
           or None if queue is not used or job_id is already running (or not found).

        :param job_id: If specified, get the position of the job
                       in the queue: None = running (probably), 0 = next to run, 1 = one job ahead, etc.
        :return: Number of tasks or None
        """
        if not self.has_queue():
            return None
        queue = self.connect_to_queue()
        if job_id is not None:
            job_ids = queue.get_job_ids()
            return job_ids.index(job_id) if job_id in job_ids else None
        else:
            return queue.qsize()

    def connect_to_queue(self, host: str = None, port: int = None):
        """Connects to the queue server and returns the queue object."""
        BaseManager.register("get_queue")
        default_host, default_port = self._queue_get_address()
        host = host or default_host
        port = port or default_port
        manager = BaseManager(address=(host, port), authkey=self._queue_get_auth_key())
        try:
            manager.connect()
            return manager.get_queue()
        except ConnectionRefusedError:
            raise ConnectionError(
                f"Could not connect to queue server at {':'.join(map(str, self._queue_get_address()))}. Is the ovo scheduler worker running?"
            )

    def create_queue_server(self, host: str = None, port: int = None):
        """Returns queue and function that runs infinitely in the server process, serving the queue to workers."""
        queue = TaskQueue()
        BaseManager.register("get_queue", callable=lambda: queue)
        default_host, default_port = self._queue_get_address()
        host = host or default_host
        port = port or default_port
        manager = BaseManager(address=(host, port), authkey=self._queue_get_auth_key())
        server = manager.get_server()
        print(f"Queue server starting on {host}:{port}")
        return queue, server.serve_forever

    def queue_worker(self, queue, queue_id):
        """Function that runs infinitely in the worker process, processing jobs from the queue."""
        print(f"Worker {queue_id} connected to queue")
        while True:
            item = queue.get()
            if item is None:
                # Shutdown signal
                print(f"Worker {queue_id} shutting down")
                return
            if not isinstance(item, tuple) or len(item) != 2:
                print(f"Worker {queue_id} received invalid task: {item} ({type(item)})")
                continue
            job_id, task = item
            print(f"Worker {queue_id} executing task: {task}")
            try:
                start_time = time.time()
                self.queue_run_task(job_id, task)
                seconds = time.time() - start_time
                print(f"Worker {queue_id} finished command in {int(seconds) // 60} minutes {seconds % 60:.0f} seconds")
            except Exception as e:
                traceback.print_exc()
                print(f"Worker {queue_id} encountered error: {e}")

    def queue_run_task(self, job_id: str, task: Any):
        """Execute a single task from the queue synchronously - executed in the worker loop"""
        raise NotImplementedError()

    def _queue_get_auth_key(self):
        return os.environ.get("OVO_WORKER_AUTHKEY", "no_key").encode("utf-8")


class TaskQueue:
    def __init__(self):
        self.q = Queue()

    def put(self, item):
        return self.q.put(item)

    def get(self):
        return self.q.get()

    def get_job_ids(self):
        with self.q.mutex:
            return [item[0] for item in self.q.queue]



import streamlit as st

from ovo import db
from ovo.app.components.custom_elements import confirm_download_button
from ovo.app.components.descriptor_explorer import descriptor_explorer
from ovo.app.components.descriptor_job_components import refresh_descriptors
from ovo.app.components.descriptor_table import descriptor_table
from ovo.app.components.descriptor_tiles import descriptor_overview_tiles
from ovo.app.components.navigation import design_navigation_selector
from ovo.app.utils.cached_db import (
    get_cached_pools,
    get_cached_design_ids,
    get_cached_available_descriptors,
)
from ovo.app.utils.protein_qc_plots import source_selectbox
from ovo.core.database import Pool, Design, NumericGlobalDescriptor
from ovo.core.database.descriptors_proteinqc import PROTEINQC_MAIN_DESCRIPTORS
from ovo.core.database.models_proteinqc import ProteinQCWorkflow, PROTEINQC_TOOLS
from ovo.core.logic.descriptor_logic import (
    submit_descriptor_workflow,
    get_wide_descriptor_table,
    export_proteinqc_excel,
)
from ovo.core.logic.proteinqc_logic import get_available_schedulers


@st.fragment
def proteinqc_fragment(pool_ids: list[str], design_ids: list[str] | None = None):
    if design_ids is None:
        # design_ids not explicitly passed, use all accepted designs in the selected pools
        design_ids = get_cached_design_ids(pool_ids=pool_ids, accepted=True)
        if not design_ids:
            st.write(
                "No accepted designs in the selected "
                + ("pools" if len(pool_ids) > 1 else "pool")
                + ". Please mark some designs as accepted in the **Jobs** page."
            )
            return
    else:
        if not design_ids:
            # Empty list explicitly passed to design_ids
            st.write("No designs selected")
            return

    st.header(f"🔎 ProteinQC | {len(design_ids):,} {'design' if len(design_ids) == 1 else 'designs'}")

    if st.button(
        "Submit ProteinQC",
        type="primary",
        key="submit_full_proteinqc_btn",
        help="Submit full ProteinQC for all designs",
    ):
        submit_proteinqc_dialog(pool_ids, design_ids)

    refresh_descriptors(
        design_ids=design_ids,
        workflow_names=[ProteinQCWorkflow.name],
    )

    st.subheader("Descriptors table")
    descriptors_by_key = get_cached_available_descriptors(design_ids)
    numeric_descriptors_by_key = {
        d_key: d for d_key, d in descriptors_by_key.items() if type(d) is NumericGlobalDescriptor
    }
    descriptors_df = get_wide_descriptor_table(
        design_ids=design_ids, descriptor_keys=descriptors_by_key.keys(), nested=True
    )

    descriptors = [d for d in PROTEINQC_MAIN_DESCRIPTORS if d.key in descriptors_by_key]

    descriptor_table(design_ids, descriptors_df, descriptors)

    if st.button("Download ProteinQC table", key="prepare_proteinqc"):
        with st.spinner("Preparing descriptor table..."):
            excel_bytes = export_proteinqc_excel(design_ids)
        confirm_download_button(
            data=excel_bytes.getvalue(),
            file_name=f"ProteinQC_{'_'.join(pool_ids)}_{len(design_ids)}_designs.xlsx",
            mime="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
            key=f"download_proteinqc",
        )

    st.subheader("Results")

    # TODO rank by number of flags, pass fmt to navigation selector
    ranked_ids = descriptors_df.index.tolist()
    left, right = st.columns([3, 1], vertical_alignment="bottom")
    with left:
        design_id = design_navigation_selector(ranked_ids, allow_all=True)
    with right:
        histogram_source = source_selectbox(st.query_params.get("ref"), key_prefix="proteinqc_reference")
        st.query_params["ref"] = histogram_source

    descriptor_overview_tiles(
        descriptors_df, descriptors_by_key, design_id=design_id, histogram_source=histogram_source
    )

    st.subheader("Descriptor explorer")
    descriptor_explorer(descriptors_df, numeric_descriptors_by_key, single_design=False)


@st.fragment
@st.dialog("ProteinQC submission", width="large")
def submit_proteinqc_dialog(pool_ids: list[str], design_ids: list[str]):
    num_designs = len(design_ids)
    st.write(f"""Submit ProteinQC for {num_designs:,} {"design" if num_designs == 1 else "designs"}""")
    tools = st.multiselect(
        "Select ProteinQC tools",
        options=PROTEINQC_TOOLS,
        default=PROTEINQC_TOOLS,
        format_func=lambda x: x.name,
        key="proteinqc_tools_selectbox",
    )
    tool_keys = [tool.tool_key for tool in tools]
    chains = st.text_input(
        "Chain(s) to analyze",
        value="A",
        key="proteinqc_chains_input",
    )
    chains = chains.replace(" ", "").replace(",", "")

    schedulers = get_available_schedulers(tools)

    if not schedulers:
        st.warning("No schedulers available for the selected tools.")
        return

    scheduler_key = st.selectbox(
        "Scheduler",
        options=list(schedulers.keys()),
        format_func=lambda x: schedulers[x].name,
        key="proteinqc_scheduler_selectbox",
    )

    if st.button("Submit", key="submit_proteinqc_btn", type="primary"):
        st.write("Submitting job... 🚀")

        workflow = ProteinQCWorkflow(tools=tool_keys, chains=list(chains), design_ids=design_ids)
        submit_descriptor_workflow(workflow, scheduler_key, st.session_state.project.id)
        st.rerun()



from typing import Type, Sequence

from ovo.core.database.base_db import DBEngine, T
from ovo.core.database.models import Base


CLEAR_FUNCTIONS = {}


def clear_when_modified(*models):
    assert models, "At least one model must be provided"
    # NOTE this can sometimes fail locally due to live reload, restarting the app will solve the issue
    assert issubclass(models[0], Base), (
        f"Expected model, got {type(models[0])} - usage: clear_when_modified(Project, Design) - do not wrap in list"
    )

    def decorator(func):
        CLEAR_FUNCTIONS[func.__name__] = (func, models)
        return func

    return decorator


class CacheClearingEngine(DBEngine):
    def save(self, obj: Base):
        if isinstance(obj, list):
            raise ValueError("Use save_all() for saving multiple objects, got list in save()")
        self._clear_all_cache(type(obj))

    def save_all(self, objs: Sequence[Base]):
        for type_ in set(type(obj) for obj in objs):
            self._clear_all_cache(type_)

    def remove(self, model: Type[T], *id_args, **kwargs):
        self._clear_all_cache(model)

    def save_value(self, model: Type[T], column: str, value, **kwargs):
        self._clear_all_cache(model)

    def _clear_all_cache(self, model_to_clear: Type[T]):
        for func_name, (func, models) in CLEAR_FUNCTIONS.items():
            for model in models:
                if issubclass(model_to_clear, model):
                    if not hasattr(func, "clear"):
                        raise ValueError(
                            f"Incorrect decorator order: "
                            f"Use @clear_when_modified before @st.cache_data - missing clear method on {func}"
                        )
                    func.clear()



import os
import platform
import shutil
import subprocess
import tempfile

from typing import List


# def protein_sol(event, context):
#     sequences = event.get('sequences')
#     if not sequences:
#         return {
#             'statusCode': 400,
#             'body': json.dumps(
#                 {
#                     'status': 'Error'
#                 }
#             )
#         }

#     predictions = call_protein_sol(sequences)

#     return {
#         'statusCode': 200,
#         'body': json.dumps(
#             {'predictions': predictions}
#         )
#     }


def protein_sol(sequences: List[str]):
    if not sequences:
        raise ValueError("No sequences provided")

    predictions = call_protein_sol(sequences)
    return predictions


def call_protein_sol(sequences):
    with tempfile.TemporaryDirectory() as tmpdirname:
        copy_script_files(tmpdirname)
        process_sequences(sequences, tmpdirname)
        return parse_predictions_file(os.path.join(tmpdirname, "seq_prediction.txt"))


def copy_script_files(dest):
    protein_sol_path = "./lib/protein-sol-sequence-prediction-software"
    script_filenames = [
        "multiple_prediction_wrapper_export.sh",
        "fasta_seq_reformat_export.pl",
        "seq_compositions_perc_pipeline_export.pl",
        "server_prediction_seq_export.pl",
        "seq_props_ALL_export.pl",
        "profiles_gather_export.pl",
        "ss_propensities.txt",
        "seq_reference_data.txt",
    ]

    for filename in script_filenames:
        shutil.copyfile(os.path.join(protein_sol_path, filename), os.path.join(dest, filename))

    os.chmod(os.path.join(dest, "multiple_prediction_wrapper_export.sh"), 0o775)


def process_sequences(sequences, directory):
    with tempfile.NamedTemporaryFile(mode="w+") as sequence_file:
        for i, sequence in enumerate(sequences):
            sequence_file.write(f">header{i}\n{sequence}\n")
        sequence_file.seek(0)
        if platform.system() == "Windows":
            subprocess.run(
                [
                    "C:\Program Files\Git\git-bash.exe",
                    os.path.join(directory, "multiple_prediction_wrapper_export.sh"),
                    sequence_file.name,
                ],
                cwd=directory,
            )
        else:
            subprocess.run(["./multiple_prediction_wrapper_export.sh", sequence_file.name], cwd=directory)


def parse_predictions_file(path):
    with open(path) as prediction_file:
        parser = SeqPredictionParser(prediction_file)
        predictions = parser.parse()
        return predictions


class SeqCompositionsParser:
    def __init__(self, file):
        self.lines = file.read().splitlines()
        self._current_line_number = 15
        self.headers = None

    def parse(self):
        headers = self.parse_headers()
        compositions = self.parse_compositions(headers)
        sums = self.parse_sums(self._current_line_number)
        return (compositions, sums)

    def parse_headers(self):
        headers_strings = (self.lines[4], self.lines[6], self.lines[8], self.lines[10], self.lines[12])
        self.headers = [[x.strip() for x in header_line] for header_line in [x.split(",") for x in headers_strings]]
        return self.headers

    def parse_compositions(self, headers):
        res = []
        while not self.lines[self._current_line_number].startswith("SUMS"):
            data = [x.split(",") for x in self.lines[self._current_line_number : self._current_line_number + 5]]
            res_single = {}
            for j, data_line in enumerate(data):
                data_line_stripped = [x.strip() for x in data_line]
                res_single[headers[j][0]] = dict(zip(headers[j][1:], data_line_stripped[1:]))
            res.append(res_single)
            self._current_line_number += 6
        return res

    def parse_sums(self, line_number):
        sums = {}
        sums["K-R"], sums["D-E"] = self.lines[line_number].split(" = ")[1].split()
        sums["K"], sums["R"], sums["D"], sums["E"] = self.lines[line_number + 2].split(" = ")[1].split()
        return sums


class SeqPredictionParser:
    def __init__(self, file):
        self.lines = file.read().splitlines()
        self._current_line_number = 14

    def parse(self):
        headers = self.parse_headers()
        return self.parse_prediction(headers)

    def parse_headers(self):
        headers_strings = self.lines[10:13]
        self.headers = [x.split(",") for x in headers_strings]
        return self.headers

    def parse_prediction(self, headers):
        res = []

        while self._current_line_number < len(self.lines):
            data = [x.split(",") for x in self.lines[self._current_line_number : self._current_line_number + 3]]
            res_single = {}

            predictions = [x.strip() for x in data[0]]
            res_single.update(zip(headers[0][2:], [float(x) for x in predictions[2:]]))

            deviations = [x.strip() for x in data[2]]
            res_single["deviations"] = dict(zip(headers[2][2:], [float(x) for x in deviations[2:]]))

            self._current_line_number += 3

            for i in range(2, 4):
                line_to_process = self.lines[self._current_line_number + i].split()[-1].split(",")
                res_single[line_to_process[0]] = ([None] * 10) + [float(x) for x in line_to_process[2:]] + ([None] * 10)

            res.append(res_single)
            self._current_line_number += 5

        return res


if __name__ == "__main__":
    protein_sol(["ASDASD"])



import json
import os


def is_running_in_streamlit():
    try:
        import streamlit
    except ImportError:
        return False
    from streamlit.runtime.scriptrunner import get_script_run_ctx

    return get_script_run_ctx(suppress_warning=True) is not None


def get_posit_user_info():
    import streamlit

    headers = streamlit.context.headers
    if not headers:
        return None
    user_info_json = headers.get("Rstudio-Connect-Credentials")
    if user_info_json is None:
        return None
    return json.loads(user_info_json)


def get_username() -> str:
    if is_running_in_streamlit():
        import streamlit as st

        if hasattr(st.user, "is_logged_in") and st.user.is_logged_in:
            return st.user["email"]
        user_info = get_posit_user_info()
        if user_info is not None:
            return user_info.get("user", "unknown")
    # get the current user of the computer (for local development)
    return os.environ.get("USER", "unknown")



from typing import Sequence, Self, Any
import pandas as pd
from sqlalchemy.orm.decl_api import DCTransformDeclarative


class DBProxy(DCTransformDeclarative):
    """Proxy class for DBEngine to enable db.Design.select() instead of db.select(Design)

    It provides a subset of methods of DBEngine (methods that operate on a single model class).
    """

    def _db(self):
        from ovo import db

        return db

    def remove(cls, *id_args, **kwargs):
        return cls._db().remove(cls, *id_args, **kwargs)

    def save_value(cls, column: str, value, **kwargs):
        return cls._db().save_value(cls, column, value, **kwargs)

    def select(cls, limit: int = None, order_by=None, **kwargs) -> Sequence[Self]:
        return cls._db().select(cls, limit=limit, order_by=order_by, **kwargs)

    def count(cls, **kwargs) -> int:
        return cls._db().count(cls, **kwargs)

    def count_distinct(cls, field="id", group_by=None, **kwargs) -> int | dict[Any, int]:
        return cls._db().count_distinct(cls, field=field, group_by=group_by, **kwargs)

    def get(cls, *id_args, **kwargs) -> Self:
        return cls._db().get(cls, *id_args, **kwargs)

    def get_value(cls, column: str, *id_args, raw=False, **kwargs):
        return cls._db().get_value(cls, column, *id_args, raw=raw, **kwargs)

    def select_values(cls, column: str, order_by=None, **kwargs) -> list:
        return cls._db().select_values(cls, column, order_by=order_by, **kwargs)

    def select_dict(cls, key_column: str, value_column: str, order_by=None, **kwargs) -> dict:
        return cls._db().select_dict(cls, key_column, value_column, order_by=order_by, **kwargs)

    def select_unique_values(cls, column: str, **kwargs) -> set:
        return cls._db().select_unique_values(cls, column, **kwargs)

    def select_dataframe(cls, index_col: str = "id", order_by=None, limit=None, **kwargs) -> pd.DataFrame:
        return cls._db().select_dataframe(cls, index_col=index_col, order_by=order_by, limit=limit, **kwargs)



from io import BytesIO
from typing import Any
import pandas as pd


def write_sheets(df_dict: dict[str, pd.DataFrame], fd_or_path: BytesIO | str, header_format: dict | None = "default"):
    writer = pd.ExcelWriter(fd_or_path, engine="xlsxwriter")

    sheet_names = shorten_sheet_names([sanitize_excel_sheet_name(n) for n in df_dict.keys()])
    for sheet_name, df in zip(sheet_names, df_dict.values()):
        has_index_name = any(df.index.names)
        write_sheet(df, writer, sheet_name=sheet_name, index=has_index_name, header_format=header_format)

    return writer


def write_sheet(
    df: pd.DataFrame,
    writer: pd.ExcelWriter,
    sheet_name: str = "Sheet1",
    index: bool = True,
    header_format: dict = "default",
    **kwargs: Any,
) -> None:
    """
    Write df as an excel file to ExcelWriter, roughly similar to `df.to_excel` except that it handles
    `df` with MultiIndex columns and `index=False`.
    """

    # Avoid the "NotImplementedError: Writing to Excel with MultiIndex columns"
    # exception by temporarily changing the columns to a single-level index
    columns: pd.Index = df.columns
    df.columns = range(len(df.columns))
    df.to_excel(
        writer,
        startrow=columns.nlevels,
        header=False,
        sheet_name=sheet_name,
        index=index,
        **kwargs,
    )
    df.columns = columns

    # Get the xlsxwriter workbook and worksheet objects.
    book = writer.book
    sheet = writer.sheets[sheet_name]

    # Add a header format.
    if header_format is not None:
        header_format = book.add_format(
            {"bold": True, "text_wrap": True, "valign": "top", "bottom": 1}
            if header_format == "default"
            else header_format
        )

    col_offset = df.index.nlevels if index else 0

    # Write the column headers with the defined format.
    for level in range(columns.nlevels):
        names = columns.get_level_values(level)
        prev_name = None
        last_name_change_i = 0
        for i, name in enumerate(names):
            if prev_name == name:
                next_name = names[i + 1] if len(names) > i + 1 else None
                if next_name != name:
                    sheet.merge_range(
                        level,
                        last_name_change_i + col_offset,
                        level,
                        i + col_offset,
                        name,
                        header_format,
                    )
                continue
            if not pd.isna(name):
                sheet.write(level, i + col_offset, name, header_format)
            last_name_change_i = i
            prev_name = name

    if index and any(df.index.names):
        for i, name in enumerate(df.index.names):
            sheet.write(df.columns.nlevels - 1, i, name or "", header_format)


def sanitize_excel_sheet_name(name: str) -> str:
    sanitized = name
    for char in "[]:*?/\\":
        sanitized = sanitized.replace(char, "_")
    return sanitized


def shorten_sheet_names(names_orig: list[str], max_length: int = 27, max_iter: int = 25) -> list[str]:
    suffixes = [0] * len(names_orig)
    names_final = [name[:max_length] for name in names_orig]
    iteration = 0
    while len(set(names_final)) < len(names_final) and iteration < max_iter:
        iteration += 1
        used = {}
        for i, name in enumerate(names_final):
            if names_final.count(name) > 1:
                number = used.get(name, 0) + 1
                used[name] = number
                suffixes[i] = number
        for i in range(len(names_orig)):
            suffix = f" ({str(suffixes[i])})" if suffixes[i] > 0 else ""
            if max_length - len(suffix) < 0:
                raise ValueError(f"Suffix of sheet name {suffix} is too long!")
            if len(names_orig[i]) > max_length - len(suffix):
                suffix = "..." + suffix
            names_final[i] = f"{names_orig[i][: max_length - len(suffix)]}{suffix}"
    if len(set(names_final)) < len(names_final):
        collisions = set([name for name in names_final if names_final.count(name) > 1])
        raise ValueError(f"The colliding sheet names could not be resolved. Collisions: {collisions}")
    return names_final



import os
import sys
import argparse
import logging
import torch
import time
import re
import json
import numpy as np

from glob import glob
from Bio import SeqIO, PDB
from Bio.PDB import PDBParser
from Bio.SeqUtils import seq1
from io import StringIO
from dataclasses import dataclass

from transformers import EsmForProteinFolding
from transformers.models.esm.openfold_utils import residue_constants
from transformers.models.esm.modeling_esmfold import collate_dense_tensors


@dataclass
class InputSequence:
    id: str
    sequence: str
    original_pdb_path: str | None = None


def get_sequence_from_pdb_file(file_path: str, chain: str = "A") -> InputSequence:
    """
    Get the sequence of a structure from a pdb file.
    """
    pdb_parser = PDBParser(QUIET=True)
    seq_id = re.sub(r"\.pdb$", "", os.path.basename(file_path))
    with open(file_path) as f:
        parsed_structure = pdb_parser.get_structure("", f)
        sequence = []
        for residue in parsed_structure[0][chain]:
            if residue.has_id("CA"):
                sequence.append(seq1(residue.get_resname()))
    return InputSequence(id=seq_id, sequence="".join(sequence), original_pdb_path=file_path)


def get_sequence_from_fasta_file(file_path: str) -> InputSequence:
    """
    Get the first sequence from a FASTA file.
    """
    seq_id = os.path.splitext(os.path.basename(file_path))[0]
    with open(file_path) as f:
        for record in SeqIO.parse(f, "fasta"):
            return InputSequence(id=seq_id, sequence=str(record.seq))
    raise ValueError(f"No valid sequences found in {file_path}")


def get_sequences_from_fasta_file(file_path: str) -> list[InputSequence]:
    """
    Get all sequences from a FASTA file.
    """
    input_sequences = []
    with open(file_path) as f:
        for record in SeqIO.parse(f, "fasta"):
            input_sequences.append(InputSequence(id=record.id, sequence=str(record.seq)))
    if len(input_sequences) == 0:
        raise ValueError(f"No valid sequences found in {file_path}")
    return input_sequences


def setup_logger(output_dir: str) -> logging.Logger:
    """Set up logging to both file and console."""
    os.makedirs(output_dir, exist_ok=True)
    log_file = os.path.join(output_dir, "logs.txt")
    open(log_file, "w").close()
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    file_handler = logging.FileHandler(log_file)
    file_handler.setLevel(logging.INFO)
    logger.addHandler(file_handler)
    return logger


@torch.no_grad()
def infer(
    model: EsmForProteinFolding,
    seqs: str | list[str],
    num_recycles: int,
    position_ids: torch.Tensor | None = None,
) -> dict:
    # Returns the raw outputs of the model given an input sequence.
    if isinstance(seqs, str):
        seqs = [seqs]
    device = next(model.parameters()).device
    aatype = collate_dense_tensors(
        [
            torch.from_numpy(
                residue_constants.sequence_to_onehot(
                    sequence=seq,
                    mapping=residue_constants.restype_order_with_x,
                    map_unknown_to_x=True,
                )
            )
            .to(device)
            .argmax(dim=1)
            for seq in seqs
        ]
    )  # B=1 x L
    mask = collate_dense_tensors([aatype.new_ones(len(seq)) for seq in seqs])
    position_ids = (
        torch.arange(aatype.shape[1], device=device).expand(len(seqs), -1)
        if position_ids is None
        else position_ids.to(device)
    )
    if position_ids.ndim == 1:
        position_ids = position_ids.unsqueeze(0)
    return model(
        aatype,
        mask,
        position_ids=position_ids,
        num_recycles=num_recycles,
    )


def get_atom_coordinates(
    structure: PDB.Structure.Structure,
    chain_id: str | None,
    residues: list[int] | None,
    all_atom: bool = False,
    model_index: int = 0,
) -> tuple[list[dict[str, np.ndarray]], list[PDB.Residue.Residue]]:
    coords = []
    res_final = []

    model = structure[model_index]

    for chain in model:
        if chain_id and chain.id != chain_id:
            continue
        for residue in chain:
            if residues and residue.id[1] not in residues:
                continue
            coords_dict = {}
            for atom in residue:
                if all_atom or atom.id == "CA":
                    coords_dict[atom.id] = atom.coord
            if coords_dict:
                coords.append(coords_dict)
                res_final.append(residue)
    return coords, res_final


def align_multiple_proteins_pdb(
    pdb_strs: list[str], chain_residue_mappings: list[list[tuple[str, list[int] | None]] | None], all_atom: bool = False
) -> float:
    """Aligns multiple protein structures based on their atoms (CA or all).

    :param pdb_strs: list of PDB strings
    :param chain_residue_mappings: list of lists of tuples with chain ID and residues to align,
                                   if None provided, then whole chain/structure is aligned
    :param all_atom: if True, align using all atoms from matched residues (not just CA atoms)
    """
    assert len(pdb_strs) == len(chain_residue_mappings), (
        f"Expected same number of structures and chain residue mappings, got {len(pdb_strs)} != {len(chain_residue_mappings)}"
    )

    parser = PDB.PDBParser(QUIET=True)
    structures = [parser.get_structure(f"Protein{i + 1}", StringIO(pdb_str)) for i, pdb_str in enumerate(pdb_strs)]
    coords_list = []
    residues_list = []

    for i, (structure, mappings) in enumerate(zip(structures, chain_residue_mappings)):
        structure_coords = []
        structure_residues = []

        if not mappings:
            coords, res_final = get_atom_coordinates(structure, None, None, all_atom=all_atom)
            if coords:
                coords_list.append(coords)
                residues_list.append(res_final)
            else:
                raise ValueError(f"No atoms found in structure {i + 1}")
        else:
            for chain_id, residues in mappings:
                coords, res_final = get_atom_coordinates(structure, chain_id, residues, all_atom=all_atom)
                if coords:
                    structure_coords.extend(coords)
                    structure_residues.extend(res_final)
                else:
                    raise ValueError(
                        f"No atoms found in structure {i + 1} for chain {chain_id} and residues {residues}"
                    )

            coords_list.append(structure_coords)
            residues_list.append(structure_residues)

    for residues in residues_list:
        assert len(residues) == len(residues_list[0]), (
            f"Got different number of residues in structures: {len(residues)} != {len(residues_list[0])}"
        )

    super_imposer = PDB.Superimposer()

    for i in range(1, len(structures)):
        ref_atoms = []
        mod_atoms = []
        for atom_id, (ref_coords, mod_coords) in enumerate(zip(coords_list[0], coords_list[i])):
            shared_atom_names = sorted(set(ref_coords.keys()) & set(mod_coords.keys()))
            for atom_name in shared_atom_names:
                atom = atom_name[0]
                ref_atoms.append(PDB.Atom.Atom("X", ref_coords[atom_name], 1.0, 1.0, " ", "X", atom_id, atom))
                mod_atoms.append(PDB.Atom.Atom("X", mod_coords[atom_name], 1.0, 1.0, " ", "X", atom_id, atom))
        super_imposer.set_atoms(ref_atoms, mod_atoms)

    return super_imposer.rms


def main(args):
    logger = setup_logger(args.output_dir)

    assert 0 <= args.num_recycles <= 4, "Number of recycles must be between 0 and 4"

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    logger.info(f"Using device: {device}")

    # Load model
    model = EsmForProteinFolding.from_pretrained(args.esmfold_model_path)
    model = model.eval()
    model = model.to(device)
    if args.fp16 and device == torch.device("cuda"):  # 16bit only for GPU
        logger.info("Using mixed precision (BFloat16) for inference")
        model = model.to(torch.bfloat16)
    logger.info("Model loaded successfully.")

    # Load input sequences
    input_sequences: list[InputSequence]
    if os.path.isdir(args.input_path):
        paths_pdb = sorted(glob(os.path.join(args.input_path, "*.pdb")))
        paths_fasta = sorted(
            glob(os.path.join(args.input_path, "*.fasta")) + glob(os.path.join(args.input_path, "*.fa"))
        )
        assert (not paths_pdb) or (not paths_fasta), "Input directory cannot contain both PDB and FASTA files"
        if paths_pdb:
            input_sequences = [get_sequence_from_pdb_file(path, args.chain) for path in paths_pdb]
        else:
            input_sequences = [get_sequence_from_fasta_file(path) for path in paths_fasta]
    elif args.input_path.endswith((".pdb",)):
        input_sequences = [get_sequence_from_pdb_file(args.input_path, args.chain)]
    elif args.input_path.endswith((".fasta", ".fa")):
        input_sequences = get_sequences_from_fasta_file(args.input_path)
    else:
        raise ValueError(
            "Input must be a directory with PDB files, a directory with FASTA files, a PDB file, or a FASTA file"
        )

    logger.info(f"Processing {len(input_sequences)} sequences")

    os.makedirs(os.path.join(args.output_dir, args.name), exist_ok=True)

    # Run inference and save output PDBs
    predictions_metrics = []
    for input_sequence in input_sequences:
        logger.info(f"Processing {input_sequence.id}")
        inference_start_time = time.time()
        prediction = infer(model, input_sequence.sequence, num_recycles=args.num_recycles)
        logger.info(f"Inference time: {time.time() - inference_start_time:.2f} seconds")

        output_dict = {
            k: v.to(torch.float32) if v.dtype in (torch.bfloat16, torch.float16) else v for k, v in prediction.items()
        }
        prediction_pdb = model.output_to_pdb(output_dict)[0]

        with open(os.path.join(args.output_dir, args.name, f"{input_sequence.id}_{args.name}.pdb"), "w") as f:
            f.write(prediction_pdb)

        RMSD_all_atom, RMSD_backbone = None, None
        if input_sequence.original_pdb_path is not None:
            with open(input_sequence.original_pdb_path, "r") as f:
                original_pdb = f.read()
            RMSD_all_atom = align_multiple_proteins_pdb(
                pdb_strs=[original_pdb, prediction_pdb],
                chain_residue_mappings=[[("A", None)], None],
                all_atom=True,
            )
            RMSD_backbone = align_multiple_proteins_pdb(
                pdb_strs=[original_pdb, prediction_pdb],
                chain_residue_mappings=[[("A", None)], None],
                all_atom=False,
            )

        predictions_metrics.append(
            dict(
                id=input_sequence.id,
                pLDDT=(100.0 * prediction.plddt[0, :, 1].mean().item()),  # (Batch x Length x 37); CA is on index 1
                pTM=prediction.ptm.item(),
                RMSD_all_atom=RMSD_all_atom,
                RMSD_backbone=RMSD_backbone,
                pAE_mean=prediction.predicted_aligned_error.mean().item(),
            )
        )

    # Save metrics to JSONL file
    with open(os.path.join(args.output_dir, f"{args.name}.jsonl"), "w") as f:
        for metric in predictions_metrics:
            json.dump(metric, f)
            f.write("\n")
            f.flush()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--input_path",
        type=str,
        help="Path to the input directory containing FASTA files or PDB files or a single FASTA or PDB file",
    )
    parser.add_argument("--output_dir", type=str, help="Path to the output directory for PDB files")
    parser.add_argument("--name", default="esmfold", type=str, help="Filename prefix to use for PDB dir and stats file")
    parser.add_argument("--chain", type=str, default="A", help='Chain identifier for input PDB files. Default: "A"')
    parser.add_argument("--esmfold_model_path", type=str, help="Path to the ESMFold model directory")
    parser.add_argument(
        "--num_recycles",
        type=int,
        default=4,
        help="Number of recycles during inference. Default is the maximum used during training (4)",
    )
    parser.add_argument(
        "--fp16",
        action="store_true",
        help="Use mixed precision (FP16) for inference",
    )
    args = parser.parse_args()
    main(args)



import subprocess
import os
from ovo.core.utils.pdb import get_sequences_from_pdb_str
from ovo.core.utils.resources import RESOURCES_DIR
from ovo.core.utils.tests import get_test_scheduler, prepare_test_output_dir


def test_scaffold_hairpin():
    scheduler = get_test_scheduler()
    output_dir = prepare_test_output_dir(__file__)
    scheduler.run(
        "rfdiffusion-backbone",
        output_dir=output_dir,
        params=dict(
            max_memory="4G",
            input_pdb=RESOURCES_DIR / "examples/inputs/5ELI_A.pdb",
            num_designs=1,
            contig="A111-114/10/A117-119",
            run_parameters="diffuser.T=1 inference.deterministic=True",
        ),
    )
    print("Saved to:", output_dir)
    subprocess.run(["ls", "-lR", output_dir.rstrip("/") + "/"], check=True)

    # read seq per chain from PDB file
    with open(os.path.join(output_dir, "rfdiffusion/rfdiffusion_standardized_pdb/rfdiffusion_0_standardized.pdb")) as f:
        seqs = get_sequences_from_pdb_str(f.read())
    assert sorted(seqs.keys()) == ["A"]
    assert seqs["A"] == "QSLHGGGGGGGGGGEAD"



from Bio import Align


def align_protein_global(a: str, b: str) -> Align.Alignment:
    aligner = Align.PairwiseAligner(mode="global", scoring="blastp")
    alignments = aligner.align(a, b)
    return alignments[0]



from contextlib import nullcontext
from typing import Callable

import pandas as pd
import streamlit as st

from ovo import db
from ovo.core.database import Round, Design
from ovo.core.utils.formatting import get_hash_of_bytes

CURRENT_SECTION_PREFIX = "current_section_"
ROUND_IDS_QUERY_PARAM = "rounds"
DESIGN_IDS_QUERY_PARAM = "designs"


def show_prev_next_sections(key: str, title: str, sections: dict[str, callable]):
    session_key = CURRENT_SECTION_PREFIX + key
    current_index = st.session_state.get(session_key, 0)
    section_labels = list(sections.keys())

    # Top buttons with title and glow
    show_buttons(
        session_key,
        current_index,
        section_labels,
        key_suffix="top",
        title=title,
        show_title=True,
        glow=current_index == 0,
    )

    # Section content
    current_section_fragment = list(sections.values())[current_index]
    current_section_fragment()

    # Bottom buttons
    st.markdown("#")  # Empty spacer
    show_buttons(session_key, current_index, section_labels, key_suffix="bottom")


def open_first_section(key: str):
    st.session_state[CURRENT_SECTION_PREFIX + key] = 0


def show_buttons(
    session_key: str,
    current_index: int,
    section_labels: list[str],
    key_suffix: str = "",
    title: str = "",
    show_title: bool = False,
    glow: bool = False,
):
    def prev_section():
        st.session_state[session_key] = st.session_state.get(session_key, 0) - 1

    def next_section():
        st.session_state[session_key] = st.session_state.get(session_key, 0) + 1

    with st.container(key="navigation_buttons_" + key_suffix):
        left, middle, right = st.columns([6, 1, 2] if current_index == 0 else [2, 3, 2], vertical_alignment="top")
        with left:
            if current_index > 0:
                # Show back button on all but first section
                st.button(
                    f"Back to {section_labels[current_index - 1]}",
                    icon=":material/arrow_back_ios:",
                    on_click=prev_section,
                    key="back_button_" + key_suffix,
                )
            elif show_title:
                # Show title instead of back button on first section
                st.title(title)

        if show_title and 0 < current_index < len(section_labels):
            with middle:
                # Title in the middle when not on first or last section
                st.html(
                    f'<div style="text-align: center; font-size: 23px; font-weight: 600; padding-top: 5px;">'
                    f"{title} ({current_index}/{len(section_labels) - 1})"
                    f"</div>"
                )

        with right:
            with st.container(key="glow" if glow else None, horizontal=True, horizontal_alignment="right"):
                if current_index < len(section_labels) - 1:
                    # Show next button on all but last section
                    st.button(
                        f"Next to {section_labels[current_index + 1]}",
                        icon=":material/arrow_forward_ios:",
                        on_click=next_section,
                        key="next_button_" + key_suffix,
                    )
                else:
                    # Empty space to align back button to the left
                    st.markdown(" ")

        st.markdown(
            """
        <style>
        .st-key-navigation_buttons_SUFFIX {
            /* Hack to reduce the whitespace at the top of the page */
            margin-top: -20px;
            margin-bottom: 20px;
        }
        .st-key-navigation_buttons_SUFFIX button {
            padding: 10px 25px;
            min-width: 200px;
        }
        </style>
        """.replace("SUFFIX", key_suffix),
            unsafe_allow_html=True,
        )
        if glow:
            # Glow effect for next button (to emphasize you can continue)
            st.markdown(
                """
            <style>
            .st-key-glow button {
                animation: glowfade 2.7s ease-out forwards;
                box-shadow: 0 0 25px rgba(91, 163, 15, 0.5);
            }
            @keyframes glowfade {
                0%   { box-shadow: 0 0 25px rgba(91, 163, 15, 0.5); }
                100% { box-shadow: 0 0 0rem rgba(91, 163, 15, 0.0); }
            }
            </style>
            """,
                unsafe_allow_html=True,
            )


def project_round_selector(
    rounds: list[Round], key_prefix="round_pills", allow_design_input=False
) -> list[str] | tuple[list[str], list[str] | None]:
    rounds_by_id = {r.id: r for r in rounds}
    key_suffix = "_".join(rounds_by_id)
    design_ids_key = f"{key_prefix}_design_ids_{key_suffix}"
    single_key = f"{key_prefix}_single_{key_suffix}"
    multi_key = f"{key_prefix}_multi_{key_suffix}"
    single_default = [rounds[0].id]
    multi_default = rounds_by_id.keys()

    if design_ids_key not in st.session_state:
        # First time loading the page, get design IDs in query arg input
        if st.query_params.get(DESIGN_IDS_QUERY_PARAM) and allow_design_input:
            st.session_state[design_ids_key] = st.query_params[DESIGN_IDS_QUERY_PARAM].replace(",", "\n")
            single_default = ["designs"]

    if single_key not in st.session_state and st.query_params.get(ROUND_IDS_QUERY_PARAM):
        # Page loaded for first time, initialize the defaults
        query_round_ids = st.query_params[ROUND_IDS_QUERY_PARAM].split(",")
        # only use the round ids if they belong to current project (ignore when switching to different project)
        if all(round_id in rounds_by_id for round_id in query_round_ids):
            single_default = "all" if len(query_round_ids) > 1 else query_round_ids[0]
            multi_default = query_round_ids

    options = list(rounds_by_id.keys())
    if len(rounds) > 1:
        options.append("all")
    if allow_design_input:
        options.append("designs")
    round_id = st.segmented_control(
        label="Select round",
        options=options,
        key=single_key,
        selection_mode="single",
        format_func=lambda round_id: "Multiple rounds"
        if round_id == "all"
        else ("Specific designs" if round_id == "designs" else rounds_by_id[round_id].name),
        default=single_default,  # Using the most recent round as default
        label_visibility="collapsed",
    )

    if not round_id:
        st.error("Please select a round above.")
        return ([], None) if allow_design_input else []

    if round_id == "designs":
        design_ids_input = st.text_area(
            "Design IDs",
            placeholder="Paste design IDs here, separated by commas, spaces, or new lines",
            key=design_ids_key,
        )
        design_ids = design_ids_input.replace(",", " ").strip().split()
        # validate design IDs
        found_design_ids = set(db.select_values(Design, "id", id__in=design_ids))
        if found_design_ids != set(design_ids):
            missing_ids = sorted(set(design_ids) - found_design_ids)
            st.warning(f"{len(missing_ids):,} design IDs were not found: {', '.join(missing_ids)}")
            if not found_design_ids:
                st.stop()
        elif found_design_ids:
            st.success(f"Showing {len(found_design_ids):,} " + ("design" if len(found_design_ids) == 1 else "designs"))

        # remove duplicates while preserving order (dict preserves insertion order since Python 3.7)
        design_ids = list(dict.fromkeys(design_ids))

        if len(design_ids) <= 100:
            # Save design IDs to query params but avoid too long URLs
            st.query_params[DESIGN_IDS_QUERY_PARAM] = ",".join(design_ids)
            if ROUND_IDS_QUERY_PARAM in st.query_params:
                del st.query_params[ROUND_IDS_QUERY_PARAM]

        return [], design_ids

    if round_id == "all":
        # Select multiple rounds
        selected_round_ids = st.segmented_control(
            label="Select rounds",
            options=rounds_by_id.keys(),
            key=multi_key,
            selection_mode="multi",
            format_func=lambda round_id: rounds_by_id[round_id].name,
            default=multi_default,
        )

        if not selected_round_ids:
            st.error("No rounds selected. Please select at least one round above.")
            return ([], None) if allow_design_input else []
    else:
        selected_round_ids = [round_id]

    st.query_params[ROUND_IDS_QUERY_PARAM] = ",".join(selected_round_ids)
    if DESIGN_IDS_QUERY_PARAM in st.query_params:
        del st.query_params[DESIGN_IDS_QUERY_PARAM]
    return (selected_round_ids, None) if allow_design_input else selected_round_ids


def pool_selector_table(pools_table: pd.DataFrame, project_id: str) -> list[str]:
    key = "_".join(pools_table["ID"])

    # Prepare table for data editor
    data_columns = pools_table.columns.tolist()

    # Initialize default selection state
    # We use table contents to define the key
    # TODO we could use {key} instead when Streamlit persists selection across table changes
    # So when table content changes, this will remember the previous selection across reruns.
    # Note that this does not solve saving user's new selection - it still gets discarded
    # because the block below doesn't even return the new selection state because streamlit resets the component.
    # User needs to click twice to select a new item in case they clicked after the table has changed.
    default_selection_key = f"pool_default_selection_{get_hash_of_bytes(pools_table.to_csv().encode())}"
    if default_selection_key in st.session_state:
        # Page is already loaded, use previous default state
        default_selection = st.session_state[default_selection_key]
    else:
        # First page load, initialize selection based on query params
        default_selection = st.query_params.get("pool_ids").split(",") if st.query_params.get("pool_ids") else []
        st.session_state[default_selection_key] = default_selection
    pools_table.insert(0, "Selected", [pool_id in default_selection for pool_id in pools_table["ID"]])
    pools_table["Job ID"] = [
        f"./jobs?pool_ids={pool_id}&project_id={project_id}" if not pd.isna(job_id) else None
        for pool_id, job_id in pools_table.set_index("ID")["Job ID"].items()
    ]

    edited = st.data_editor(
        pools_table,
        key=f"pools_{key}",
        hide_index=True,
        width="content",
        disabled=data_columns,
        column_config={
            "Selected": "",
            "Job ID": st.column_config.LinkColumn(display_text=r"Job ↑", disabled=True),
        },
    )

    if edited["Selected"].any():
        selected_pools_table = pools_table[edited["Selected"]]
        st.caption(
            f"Showing designs from {len(selected_pools_table)} selected "
            f"{'pool' if len(selected_pools_table) == 1 else 'pools'}."
        )
        st.query_params["pool_ids"] = ",".join(selected_pools_table["ID"].astype(str).tolist())
    else:
        selected_pools_table = pools_table
        st.caption("Showing designs from all pools. Use checkboxes above to select specific pools.")
        if "pool_ids" in st.query_params:
            del st.query_params["pool_ids"]

    selected_pool_ids = list(selected_pools_table["ID"])
    return selected_pool_ids


def previous_design_idx(idx):
    if idx > 0:
        idx -= 1
    return idx


def next_design_idx(idx, n):
    if idx < n:
        idx += 1
    return idx


def design_navigation_selector(
    design_ids: list[str],
    key: str = "selected_design",
    fmt: dict[str, str] | Callable | None = None,
    allow_all: bool = False,
) -> str | None:
    """
    Display navigation controls (prev/next buttons and dropdown) for browsing designs.
    Updates query params and returns the currently selected design_id.

    Args:
        design_ids: List of design IDs to navigate through
        key: Query parameter key for storing selected design (default: "selected_design")
        fmt: Optional dict or function that maps design IDs to labels for display in the selectbox
        allow_all: If True, allows selecting all designs (returns None)

    Returns:
        str: Currently selected design_id or None if all designs are selected
    """
    if not design_ids:
        return None
    if allow_all and len(design_ids) == 1:
        # simplify case when only one design is available - select the design directly
        allow_all = False
    num_designs = len(design_ids)
    if allow_all:
        design_ids = ["ALL"] + list(design_ids)
    # Get current index from query params
    if key in st.query_params and st.query_params[key] in design_ids:
        idx = design_ids.index(st.query_params[key])
        element_key = f"{key}_selectbox_{idx}"
        if element_key in st.session_state and st.session_state[element_key] in design_ids:
            idx = design_ids.index(st.session_state[element_key])
    else:
        idx = 0
    element_key = f"{key}_selectbox_{idx}"

    with st.container(horizontal=True, gap="small", vertical_alignment="center"):
        if st.button(
            ":material/arrow_back_ios:",
            key=f"previous_design_btn_{key}",
            width="content",
            disabled=idx == 0,
        ):
            idx = previous_design_idx(idx)
            element_key = f"{key}_selectbox_{idx}"

        count_container = st.container(width=75, horizontal_alignment="center")

        if st.button(
            ":material/arrow_forward_ios:",
            key=f"next_design_btn_{key}",
            width="content",
            disabled=idx == len(design_ids) - 1,
        ):
            idx = next_design_idx(idx, len(design_ids) - 1)
            element_key = f"{key}_selectbox_{idx}"

        with count_container:
            # Write count later so that the idx is at the most recent value
            curr = "All" if design_ids[idx] == "ALL" else (idx if allow_all else idx + 1)
            st.html(
                f'<div style="text-align: center">{curr} / {num_designs:,}</div>',
            )

        def format_func(design_id):
            if design_id == "ALL":
                return f"All {num_designs:,} designs"
            if fmt:
                label = fmt.get(design_id, "") if isinstance(fmt, dict) else fmt(design_id)
            else:
                label = ""
            return f"{design_id}" + (f" | {label}" if label else "")

        design_id = st.selectbox(
            "Select a design",
            options=design_ids,
            label_visibility="collapsed",
            key=element_key,  # we add idx to force re-creating the component when idx changes
            index=idx,
            format_func=format_func,
            width="stretch",
        )
        st.query_params[key] = design_id

    return None if design_id == "ALL" else design_id



import re

from ovo import config, local_scheduler, storage
from ovo.app.components.history_components import history_dropdown_component
from ovo.app.components.input_components import pdb_input_component, sequence_selection_fragment, initialize_workflow
from ovo.app.components.molstar_custom_component import ContigsParser, molstar_custom_component, StructureVisualization
from ovo.app.components.navigation import show_prev_next_sections
from ovo.app.components.preview_components import parameters_binder_preview_component, visualize_rfdiffusion_preview
from ovo.app.components.scheduler_components import wait_with_statusbar
from ovo.app.components.submission_components import (
    pool_submission_inputs,
    show_rfdiffusion_advanced_settings,
    review_workflow_submission,
    show_rfdiffusion_binder_seq_design_inputs,
)
from ovo.app.components.trim_components import parameters_trim_structure_component, trimmed_structure_visualizer

from ovo.app.pages import jobs_page, designs_page
from ovo.app.utils.page_init import initialize_page
from ovo.core.auth import get_username
from ovo.core.database import descriptors_rfdiffusion, descriptors_refolding
from ovo.core.database.models_rfdiffusion import (
    RFdiffusionBinderDesignWorkflow,
    RFdiffusionWorkflow,
    MODEL_WEIGHTS_BINDER,
)
from ovo.core.logic.design_logic_rfdiffusion import submit_rfdiffusion_preview
from ovo.core.utils.formatting import get_hashed_path_for_bytes
from ovo.core.utils.pdb import check_rfdiffusion_input
from ovo.core.utils.residue_selection import from_contig_to_residues
from ovo.core.utils.residue_selection import get_chains_and_contigs
import streamlit as st


@st.fragment
def intro_step():
    # Initialize the workflow object in session state
    initialize_workflow(__file__, RFdiffusionBinderDesignWorkflow.name)

    with st.container(width=850):
        st.markdown(
            f"""
This workflow enables designing new binders to your input structure. It implements and extends the end-to-end workflow 
for binder design from RFdiffusion<sup>[1]</sup>.
            
Use case: designing linear peptide or miniprotein binders to a desired site on a target protein.
            
This workflow combines RFdiffusion<sup>[1]</sup> for de novo backbone design, ProteinMPNN PyRosetta FastRelax
protocol<sup>[2]</sup> for sequence design and side-chain rotamer prediction, and AlphaFold2<sup>[3]</sup> for independent
validation.
        
- **Input**
    - Target protein structure (PDB file or UniProt ID)
- **Main parameters**
    - Target chain ID and trimming boundary
    - Binder length range (peptide to miniprotein)
    - Number of backbone designs and relax steps
    - (Optional) Hotspot residues to select preferred binding site
    - (Optional) ProteinMPNN amino acid bias to enrich or deplete specific amino acids in the binder
- **Output**
    - Designs — binder sequences and structures in complex with target protein
    - Design descriptors
        - Refolding metrics
        - Rosetta descriptors
        - Backbone metrics
        - Sequence composition
            

### Features

After inputting your target structure, you can choose the desired binding site by selecting "hotspot" residues, 
or you can leave it blank to allow binding anywhere on the target surface.
Next, you'll be asked to trim the protein chain. Runtime scales quadratically with the
number of residues in the input structure, so trimming the protein to relevant regions (e.g. around a hotspot, if specified) 
will speed up the design process. After specifying the desired binder length range, you can optionally generate 
a quick RFdiffusion preview of the design with reduced number of timesteps to verify your inputs.
Finally, you will set the number of designs to generate and optional advanced settings and submit the workflow.

Once the workflow has finished, all designs are saved to the OVO *storage* and registered in the OVO *database*.
You will be able to view all designs and their descriptors in the {jobs_page.title} results page and adjust thresholds to select *accepted designs*.
The accepted designs (usually only around 0.1% to 1% of all generated designs in typical RFdiffusion workflows) can be further 
analyzed in the {designs_page.title} page. For example, you can run ProteinQC to compute descriptors of predicted solubility, electrostatic and hydrophobic patch area, and more.
    
#### Backbone metrics
- Computed for the RFdiffusion backbone design (before sequence design)
- **Secondary structure composition**:
    - Percentages of helix, sheet, loop (the remaining residues); computed using PyDSSP
- Backbone-backbone contacts (Cα atoms within 8&nbsp;Å):
    - **{descriptors_rfdiffusion.N_CONTACTS_TO_INTERFACE.name}** — Number of pairwise contacts between all residues of binder and target
    - **{descriptors_rfdiffusion.N_CONTACTS_TO_HOTSPOTS.name}** — Number of pairwise contacts between the binder backbone and hotspot residues
    - **{descriptors_rfdiffusion.N_HOTSPOTS_ON_INTERFACE.name}** — Number of hotspot residues in contact with the binder backbone

#### Sequence composition
- Percentage of individual amino acids
- Charge at pH 7.4 and 5.5, isoelectric point
- Aromaticity
- Sequence entropy, highlighting sequences with compositionally biased regions
- And more; computed using BioPython ProteinAnalysis<sup>[4]</sup>

#### Rosetta descriptors<sup>[2]</sup>
- **{descriptors_rfdiffusion.PYROSETTA_DDG.name}** — Rosetta binding energy of the complex (in Rosetta energy units). Lower (negative) values indicate better binding affinity.
- **{descriptors_rfdiffusion.PYROSETTA_CMS.name}** — Rosetta molecular surface area (in square angstroms).
- **{descriptors_rfdiffusion.PYROSETTA_SAP_SCORE.name}** — Evaluates hydrophobicity of surface exposed regions. Positive values indicate hydrophobic molecules.

#### AlphaFold2 initial guess refolding metrics<sup>[2][3][5]</sup>
- **{descriptors_refolding.AF2_PRIMARY_IPAE.name}** — Binder–target pose error (predicted) in angstroms. Predicted Aligned Error (residue–residue 
    matrix) averaged over binder→target residue values.
- **{descriptors_refolding.AF2_PRIMARY_TARGET_ALIGNED_BINDER_RMSD.name}** — Cα RMSD of the binder between RFdiffusion output and AF2 prediction (aligned on target).
  This measures how much the designed binder structure agrees with its AF2 prediction.
- **{descriptors_refolding.AF2_PRIMARY_BINDER_PAE.name}** — Binder error (predicted) including the binder–target pose, in angstroms. Predicted Aligned Error 
    <sup>[5]]</sup> (residue–residue matrix) averaged over binder→binder and binder→target residue values.

We also include AF2 binder pLDDT and pTM scores. In RFdiffusion<sup>[1]</sup> in silico success for binder design was
defined as AF2&nbsp;Binder&nbsp;pLDDT >&nbsp;80, AF2&nbsp;Interface&nbsp;PAE <&nbsp;10 and AF2&nbsp;monomer&nbsp;RMSD
<&nbsp;1.0 (we do not include this metric with the binder predicted in isolation, you could use
AF2&nbsp;Binder&nbsp;RMSD which includes the pose RMSD in the complex).
                
### Limitations

- ⚠️ **RFdiffusion workflows are computationally intensive** and typically require many designs (10,000 backbones or more) to obtain a few high-quality designs. 
     Such a workflow can take weeks of GPU time (finishing under a day thanks to parallelization). Please consider starting with a small number of designs (100-200 backbones) to verify your inputs and settings before scaling up.
- ⚠️ **Only one target chain is supported.** You may circumvent this by merging multiple chains into a single chain in your 
    input PDB file, adding residue index jumps (gap of at least 50 residues between the original chains), this influences both RFdiffusion 
    and AlphaFold2. <!--clipped to -32, 32 in alphafold2, colabdesign adds 50 between binder and target --->
- ⚠️ **High AF2 Target-aligned Binder RMSD for good designs in symmetric targets.** In case the binding site occurs multiple times in the 
    target by symmetry (e.g. a homo-n-mer)
    AF2 may predict the binder to the other occurrence of the binding site. This will cause the RMSD to be high, even
    though the design can indeed bind the correct epitope. The designs' RMSD can adopt a multimodal distribution as a
    result, where the peaks correspond to the binding site-to-binding site distances. Currently, we do not correct for
    this symmetry and when filtering designs based on AF2 Target-aligned Binder RMSD in such cases, some good designs may be discarded.
- ⚠️ **This page enables running the end-to-end workflow**. Individual steps like ProteinMPNN redesign or AlphaFold2 scoring 
    alone cannot currently be executed through the user interface.

### Methods
                    
RFdiffusion generates a backbone → PyRosetta FastRelax–ProteinMPNN protocol generates sequences and side-chain
rotamers → AlphaFold2 evaluates design by repredicting the complex, computing the confidence of the prediction, 
and by comparing the binder pose from RFdiffusion backbone to the AlphaFold2 prediction.

|             | Purpose     |  Input | Output |
|-------------|-------------|-------------|-------------|
| RFdiffusion | Backbone generation       | Trimmed target protein, hotspots (optional), binder length | Complex backbone structure |
| FastRelax–ProteinMPNN  | Sequence generation and side-chain rotamer prediction | Complex backbone structure with target sequence | Binder sequence and complex structure with side-chains |
| LigandMPNN  | Sequence generation and side-chain rotamer prediction | Backbone structure | Binder sequence and complex structure with side-chains |
| AlphaFold2  | Refolding evaluation (via structure prediction)      | Designed sequence & structure | Predicted structure |
| PyRosetta   | Additional descriptors used for filtering  |  Designed sequence & structure  | Complex binding energy, Contact molecular surface, and other descriptors  |

**Details** 
               
RFdiffusion generation
- RFdiffusion does not explicitly predict sequences for the binder; however it uses an implicit notion of sequence 
  while generating backbones. The authors note it could have been trained to co-design sequence and backbone, but they 
  found the RFdiffusion + ProteinMPNN workflow works well. Using ProteinMPNN also enables generating multiple sequences
  for the same backbone.
- Note that RFdiffusion discards all side-chain atoms of the target in the output PDB. These are reconstructed by LigandMPNN and by the AlphaFold2 prediction.

ProteinMPNN FastRelax protocol
- Successive rounds of ProteinMPNN sequence design and PyRosetta FastRelax aims to converge to a low-energy sequence–structure pair.
- Designs from each round are retained, each producing one (usually similar) sequence and structure.
- Note that we modified the protocol to perform relaxation of the final design so that all designs are relaxed and ready for downstream analysis.

AlphaFold2 protocol
- Initial guess: the complex structure (from FastRelax) is supplied to the structure module before the first recycle, 
    influencing the starting conformation. Target chain structure is always provided as template input.
- Runs in single-sequence mode (no MSA), achieving speeds of around 20 seconds for a 100aa binder and 100aa target on a low-end T4 GPU. 
  However, scales quadratically with sequence length (same as regular AlphaFold2).

### References
1. [RFdiffusion](https://www.nature.com/articles/s41586-023-06415-8), (and [Binder Design GitHub guide](https://github.com/RosettaCommons/RFdiffusion?tab=readme-ov-file#binder-design))
2. [PyRosetta FastRelax–ProteinMPNN protocol](https://www.nature.com/articles/s41467-023-38328-5)
3. [AlphaFold2](https://www.nature.com/articles/s41586-021-03819-2)
4. [BioPython ProteinAnalysis](https://biopython.org/docs/latest/api/Bio.SeqUtils.ProtParam.html#Bio.SeqUtils.ProtParam.ProteinAnalysis)
5. [ColabDesign](https://github.com/sokrypton/ColabDesign)
6. [Predicted Aligned Error (PAE)](https://www.ebi.ac.uk/training/online/courses/alphafold/inputs-and-outputs/evaluating-alphafolds-predicted-structures-using-confidence-scores/pae-a-measure-of-global-confidence-in-alphafold-predictions/)
""",
            unsafe_allow_html=True,
        )


@st.fragment
def input_step():
    workflow: RFdiffusionBinderDesignWorkflow = st.session_state.workflows[__file__]

    # Something is wrong - this object should not be None
    if not workflow.rfdiffusion_params:
        return

    left, _, right = st.columns([4, 1, 3])
    with left:
        st.subheader("Input structure")
        if new_pdb_input := pdb_input_component(workflow.input_name):
            input_name, pdb_input_bytes = new_pdb_input

            workflow.rfdiffusion_params.input_pdb = storage.store_file_str(
                pdb_input_bytes.decode(),
                f"project/{st.session_state.project.id}/inputs/{get_hashed_path_for_bytes(pdb_input_bytes)}/{input_name}.pdb",
                overwrite=False,
            )
            workflow.input_name = input_name
            workflow.selected_segments = []
            workflow.rfdiffusion_params.hotspots = None
            workflow.target_chain = None
            # TODO use presets to initialize this instead
            workflow.refolding_params.primary_test = "af2_model_1_multimer_tt_3rec"
    with right:
        history_dropdown_component(
            page_key=__file__,
            workflow_name=RFdiffusionBinderDesignWorkflow.name,
            filter_func=lambda w: not w.rfdiffusion_params.partial_diffusion,
        )

    if not workflow.rfdiffusion_params.input_pdb:
        # No input structure provided yet, stop here
        return

    st.subheader(workflow.input_name)

    molstar_custom_component(
        structures=[
            StructureVisualization(pdb=storage.read_file_str(workflow.rfdiffusion_params.input_pdb), color="chain-id")
        ],
        key="input_structure",
        width=700,
        height=400,
    )


@st.fragment
def hotspots_step():
    workflow: RFdiffusionBinderDesignWorkflow = st.session_state.workflows[__file__]

    st.subheader("Select binding hotspots")

    if not workflow.rfdiffusion_params.input_pdb:
        st.error("Please provide an input structure in the input structure step.")
        return

    sequence_selection_fragment(__file__, workflow.input_name, color="hydrophobicity", write_segments=False)


@st.fragment()
def trim_step():
    workflow: RFdiffusionBinderDesignWorkflow = st.session_state.workflows[__file__]
    st.subheader("Target chain trimming")

    if not workflow.rfdiffusion_params.input_pdb:
        st.error("Please provide an input structure in the input structure step.")
        return

    pdb_input_string = storage.read_file_str(workflow.rfdiffusion_params.input_pdb)
    workflow.chains = get_chains_and_contigs(pdb_input_string)

    # Input parameters for the preview
    parameters_trim_structure_component(workflow)

    # Check if inputs are set
    if workflow.start_res_trimmed_chain is None or workflow.end_res_trimmed_chain is None:
        return

    # Make sure that structure does not contain chain breaks without a gap in residue
    # numbering (required in refolding step)
    try:
        check_rfdiffusion_input(
            pdb_input_string, workflow.target_chain, workflow.start_res_trimmed_chain, workflow.end_res_trimmed_chain
        )
    except ValueError as e:
        st.error(str(e))
        return

    # Visualize trimmed structure
    with st.container(height=630, border=False):
        trimmed_structure_visualizer(workflow, pdb_input_string)


@st.fragment()
def preview_step():
    workflow: RFdiffusionBinderDesignWorkflow = st.session_state.workflows[__file__]
    st.subheader("Enter binder settings and preview design")

    if not workflow.rfdiffusion_params.input_pdb:
        st.error("Please provide an input structure in the input structure step.")
        return

    # Input parameters for the preview
    parameters_binder_preview_component(workflow)

    # Generate preview
    st.write("#### Generate preview")
    num_timesteps = 15
    with st.columns([2, 1])[0]:
        st.write(f"""
        Generate a quick RFdiffusion preview of the design with reduced number of timesteps 
        ({num_timesteps}/50) to verify your inputs. This step is optional.
        
        This should take 2-10 minutes depending on the length of the target and binder.
        """)

    if st.button(":material/wand_stars: Generate preview"):
        with st.spinner("Submitting RFdiffusion job..."):
            workflow.preview_job_id = submit_rfdiffusion_preview(workflow, timesteps=num_timesteps)

    # Check if needed parameters are set
    if not workflow.preview_job_id:
        return

    if workflow.rfdiffusion_params.contig is None:
        return

    # Wait until task is done
    result = wait_with_statusbar(local_scheduler, workflow.preview_job_id, label="Running RFdiffusion...")

    # Exit if job failed
    if not result:
        return

    visualize_rfdiffusion_preview(workflow, output_dir=local_scheduler.get_output_dir(workflow.preview_job_id))


@st.fragment()
def settings_step():
    st.subheader("Settings")

    workflow: RFdiffusionWorkflow = st.session_state.workflows[__file__]

    if not workflow.rfdiffusion_params.input_pdb:
        st.error("Please provide an input structure in the input structure step.")
        return

    if not workflow.rfdiffusion_params.contig:
        st.error("Contig has not been computed. Please fill in all inputs in previous step.")
        return

    pool_submission_inputs(__file__)

    with st.columns([1, 2])[0]:
        is_admin = get_username() in config.auth.admin_users
        workflow.rfdiffusion_params.num_designs = st.number_input(
            "Number of structure designs (RFdiffusion backbones)",
            min_value=1,
            max_value=config.props.rfdiffusion_backbones_limit_admin
            if is_admin
            else config.props.rfdiffusion_backbones_limit,
            value=workflow.rfdiffusion_params.num_designs,
            key="num_designs",
        )

    show_rfdiffusion_binder_seq_design_inputs(workflow)

    with st.columns([1, 2])[0]:
        workflow.rfdiffusion_params.model_weights = st.selectbox(
            "Model weights",
            help="Use 'active site' model weights to hold better selected residues specified in the contig.",
            index=MODEL_WEIGHTS_BINDER.index(workflow.rfdiffusion_params.model_weights)
            if workflow.rfdiffusion_params.model_weights
            else 0,
            key="active_site",
            options=MODEL_WEIGHTS_BINDER,
        )

    contig = st.text_input(
        "Contig",
        placeholder="A123-456/0 20-40",
        value=workflow.rfdiffusion_params.contig,
        key="input_contig",
    )

    # Step needed to update contig and (if binder design) target_chain and binder length
    if contig != workflow.rfdiffusion_params.contig and check_contig_parsed(contig, True):
        workflow.rfdiffusion_params.contig = contig

        chain_segment, workflow.binder_length = workflow.rfdiffusion_params.contig.split("/0 ")

        workflow.target_chain = chain_segment[0]

        residues = from_contig_to_residues(workflow.rfdiffusion_params.contig)
        workflow.start_res_trimmed_chain = residues[0]
        workflow.end_res_trimmed_chain = residues[-1]

    with st.columns([1, 2])[0]:
        workflow.rfdiffusion_params.hotspots = st.text_input(
            "Hotspot residues (optional)",
            placeholder="For example A123,A124,A131",
            value=workflow.rfdiffusion_params.hotspots,
            key="hotspots",
        )
        if workflow.rfdiffusion_params.hotspots:
            if not all(
                re.fullmatch("[A-Z][0-9]+", hotspot) for hotspot in workflow.rfdiffusion_params.hotspots.split(",")
            ):
                st.error("Invalid hotspots format, expected 'A123,A124,A131'")

    show_rfdiffusion_advanced_settings(workflow)


@st.fragment
def review_step():
    st.subheader("Review settings")

    review_workflow_submission(__file__)


def check_contig_parsed(contig: str | None, verbose: bool = False) -> bool:
    parser = ContigsParser()
    if not len(contig.split("/0 ")) == 2:
        if verbose:
            st.error("Contig must be in the format 'A123-456/0 20-40'")
        return False

    if contig.islower():
        st.warning("Input contigs are lowercase. Workflow can behave unexpectedly.")
    try:
        _ = parser.parse_contigs_str(contig) if contig else None
        return True
    except Exception as e:
        if verbose:
            st.error(
                f"There is an error in the contig specification. Please make sure you provide valid contigs. Error: {e}"
            )
        return False


if __name__ == "__main__":
    initialize_page(page_title="RFdiffusion Binder design")

    show_prev_next_sections(
        key=__file__,
        title="🧬 Binder design",
        sections={
            "intro": intro_step,
            "input": input_step,
            "hotspots": hotspots_step,
            "trim": trim_step,
            "preview": preview_step,
            "settings": settings_step,
            "review": review_step,
        },
    )
